# This Python file uses the following encoding: utf-8

# if__name__ == "__main__":
#     pass
# from baangt.ui.pyqt.uidesign import Ui_MainWindow
from baangt.ui.pyqt.uiDesign import Ui_MainWindow
from PyQt5 import QtWidgets, QtCore, QtGui
from PyQt5.QtCore import pyqtSlot
import os
import glob
import json
from pathlib import Path
import baangt.base.GlobalConstants as GC
from baangt.base.Utils import utils
import logging
import configparser
# import subprocess
import sys
from baangt.ui.pyqt import resources
from baangt.ui.pyqt.settingsGlobal import GlobalSettings
from baangt.ui.ImportKatalonRecorder import ImportKatalonRecorder
import pyperclip
import platform
from baangt.base.PathManagement import ManagedPaths
from uuid import uuid4
from baangt.base.FilesOpen import FilesOpen
from baangt.base.RuntimeStatistics import Statistic
from baangt.base.PathManagement import ManagedPaths
from baangt.base.DownloadFolderMonitoring import DownloadFolderMonitoring
import xlrd
from threading import Thread
from time import sleep

logger = logging.getLogger("pyC")


class PyqtKatalonUI(ImportKatalonRecorder):
    """ Subclass of ImportKatalonRecorder :
        Aim : To disable GUI created by PySimpleGui
        and initialize everything
    """
    def __init__(self, directory=None):
        self.managedPaths = ManagedPaths()
        if directory == None:
            directory = self.managedPaths.derivePathForOSAndInstallationOption()
        self.directory = directory
        self.clipboardText = ""
        self.outputText = ""
        self.window = None
        self.outputData = {}
        self.outputFormatted = []
        self.fileNameExport = None


class MainWindow(Ui_MainWindow):
    """ BaangtUI : Logic implementation file for uidesign
    """

    switch_window = QtCore.pyqtSignal(str)

    def __init__(self):
        ''' Init the super class '''
        super().__init__()
        self.lTestRun = None
        self.configContents = None
        self.__log_state = 0

    def setupUi(self, MainWindow, directory=None):
        ''' Setup the UI for super class and Implement the
        logic here we want to do with User Interface
        '''
        super().setupUi(MainWindow)
        self.managedPaths = ManagedPaths()
        if directory == None:
            directory = self.managedPaths.derivePathForOSAndInstallationOption()
        self.directory = directory
        self.configFile = None
        self.configFiles = []
        self.configContents = {}
        self.tempConfigFile = None
        self.testRunFile = None
        self.testRunFiles = []
        self.__execute_button_state = "idle"
        self.__result_file = ""
        self.__log_file = ""
        self.__open_files = 0

        # self.refreshNew()
        # self.setupBasePath(self.directory)
        self.readConfig()
        self.logSwitch.setChecked(self.__log_state)
        self.show_hide_logs()
        self.openFilesSwitch.setChecked(self.__open_files)

        self.katalonRecorder = PyqtKatalonUI(self.directory)
        # update logo and icon
        self.updateLogoAndIcon(MainWindow)

        # initialize Katalon Importer and Global Setting Page
        # Add Button Signals and Slot here
        self.browsePushButton_4.clicked.connect(self.browsePathSlot)
        # self.executePushButton.clicked(self.executeTest)

        # Setting Page actions and triggered
        self.settingsPushButton_4.clicked.connect(self.settingView)
        self.okPushButton.clicked.connect(self.saveToFile)
        self.settingComboBox_4.activated.connect(self.updateSettings)
        self.testRunComboBox_4.activated.connect(self.updateRunFile)
        self.exitPushButton.clicked.connect(self.mainPageView)
        self.AddMorePushButton.clicked.connect(self.addMore)
        self.deleteLastPushButton.clicked.connect(self.deleteLast)
        self.saveAspushButton.clicked.connect(self.saveAsNewFile)
        self.executePushButton_4.clicked.connect(self.executeButtonClicked)

        # FileOpen buttons
        self.openResultFilePushButton_4.clicked.connect(self.openResultFile)
        self.openLogFilePushButton_4.clicked.connect(self.openLogFile)
        self.openTestFilePushButton_4.clicked.connect(self.openTestFile)

        # Quit Event
        self.actionExit.triggered.connect(self.quitApplication)

        # Katalon triggered
        self.actionImport_Katalon.triggered.connect(self.show_katalon)
        self.exitPushButton_3.clicked.connect(self.exitKatalon)
        self.savePushButton_2.clicked.connect(self.saveTestCase)
        self.copyClipboard_2.clicked.connect(self.copyFromClipboard)
        self.TextIn_2.textChanged.connect(self.importClipboard)
        self.logSwitch.clicked.connect(self.show_hide_logs)
        self.openFilesSwitch.clicked.connect(self.change_openFiles_state)

        self.statistics = Statistic()

        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def close(self, event):
        try:
            self.run_process.kill()
        except:
            pass
        self.child.terminate()
        self.child.waitForFinished()
        event.accept()

    def saveInteractiveGuiConfig(self):
        """ Save Interactive Gui Config variables """
        config = configparser.ConfigParser()
        config["Default"] = {
                    "path": self.directory,
                    "testrun": self.testRunComboBox_4.currentText(),
                    "globals": self.settingComboBox_4.currentText(),
                    "logstate": self.__log_state,
                    "openfiles": self.__open_files
                    }
        with open(self.managedPaths.getOrSetIni().joinpath("baangt.ini"), "w" ) as configFile:
            config.write(configFile)

    def readConfig(self):
        """ Read existing baangt.ini file """
        config = configparser.ConfigParser()
        try:
            config.read(self.managedPaths.getOrSetIni().joinpath("baangt.ini"))
            self.directory = config["Default"]['path']
            self.testRunFile = config["Default"]['testrun']
            self.configFile  = config["Default"]['globals']
            if 'logstate' in config["Default"]:
                self.__log_state = int(config["Default"]["logstate"])
            else:
                self.__log_state = 0
            if 'openfiles' in config["Default"]:
                self.__open_files = int(config["Default"]["openfiles"])
            else:
                self.__open_files = 0
            self.setupBasePath(self.directory)
            self.readContentofGlobals()
        except Exception as e:
            print("Exception in Main readConfig. Starting with defaults", e)
            self.directory = self.managedPaths.derivePathForOSAndInstallationOption().joinpath("examples")
            if not self.directory.is_dir():
                self.directory = str(self.managedPaths.derivePathForOSAndInstallationOption())
            else:
                self.directory = str(self.directory)
            self.setupBasePath(self.directory)

    def readContentofGlobals(self):
        """ This will read the content of config file """
        configInstance = GlobalSettings.getInstance()
        configInstance.addValue(self.configFile)
        self.configContents = configInstance.config
        if not self.configContents.get('TC.' + GC.DATABASE_LINES):
            key = 'TC.' + GC.DATABASE_LINES
            self.configContents[key] = ""
        if not self.configContents.get('TC.' + GC.EXECUTION_DONTCLOSEBROWSER):
            key = 'TC.' + GC.EXECUTION_DONTCLOSEBROWSER
            self.configContents[key] = ""
        if not self.configContents.get('TC.' + GC.EXECUTION_SLOW):
            key = 'TC.' + GC.EXECUTION_SLOW
            self.configContents[key] = ""

    @QtCore.pyqtSlot()
    def show_katalon(self):
        """ Display katalon panel for Test case preparation """
        self.stackedWidget.setCurrentIndex(2)
        self.statusMessage("Katalon Studio is triggered", 1000)

    def updateLogoAndIcon(self, MainWindow):
        """ This function initialize logo and icon """
        logo_pixmap = QtGui.QPixmap(":/baangt/baangtlogo")
        logo_pixmap.scaled(300, 120, QtCore.Qt.KeepAspectRatio)
        self.logo_4.setPixmap(logo_pixmap)
        icon = QtGui.QIcon()
        icon.addPixmap(
                QtGui.QPixmap(":/baangt/baangticon"),
                QtGui.QIcon.Normal,
                QtGui.QIcon.Off
                )
        MainWindow.setWindowIcon(icon)

        self.mainPage.setStyleSheet(
                 "QLineEdit { background-color: white; \n"
                 "            color: rgb(46, 52, 54);  \n"
                 "}"
                 "QComboBox { background-color: white; \n"
                 "            color: rgb(46, 52, 54);  \n"
                 "}"
                 "QButton { color: white; \n"
                 "}"
                 )

        self.settingPage.setStyleSheet(
                 "QLineEdit { background-color: white; \n"
                 "            color: rgb(46, 52, 54);  \n"
                 "}"
                 "QComboBox { background-color: white; \n"
                 "            color: rgb(46, 52, 54);  \n"
                 "}"
                 "QLabel { font: 75 11pt 'Arial';\n"
                 "}"
                 "QButton { color: white; \n"
                 "}"
                 )
        self.katalonPage.setStyleSheet(
                 "QButton { color: white; \n"
                 "}"
                 )

        MainWindow.resize(980, 480)


    def statusMessage(self, str, duration=1000):
        """ Display status message passed in Status Bar
        Default duration (ms) = 1000 (1 sec)
        """
        self.statusbar.showMessage(str, duration)

    def getSettingsAndTestFilesInDirectory(self, dirName):
        """ Scan for *.xlsx files and *.json files and
        update the testRunComboBox and settingsComboBox items
        """
        if not dirName and not os.path.isdir(dirName):
            dirName = os.getcwd()
        # for getting back to original directory
        orig_path = os.getcwd()
        os.chdir(dirName)
        # self.testRunFiles = glob.glob("*.xlsx")
        # self.configFiles = glob.glob("global*.json")
        self.testRunFiles = []
        self.configFiles = []

        fileList = glob.glob("*.json")
        fileList.extend(glob.glob("*.xlsx"))
        if not platform.system().lower() == 'windows':
            # On MAC and LINUX there may be also upper/lower-Case versions
            fileList.extend(glob.glob("*.JSON"))
            fileList.extend(glob.glob("*.XLSX"))
        for file in fileList:
            if file[0:6].lower() == 'global':      # Global Settings for Testrun must start with global_*
                self.configFiles.append(file)
            else:
                self.testRunFiles.append(file)
            pass

        # get back to orig_dir
        os.chdir(orig_path)

        # update the combo box
        self.testRunComboBox_4.clear()
        self.settingComboBox_4.clear()
        # Also, disable Execute and Details Button
        self.executePushButton_4.setEnabled(False)
        self.settingsPushButton_4.setEnabled(False)
        # Add files in Combo Box
        self.testRunComboBox_4.addItems(
                        sorted(self.testRunFiles, key=lambda x: x.lower())
                               )
        self.settingComboBox_4.addItems(
                        sorted(self.configFiles, key=lambda x: x.lower())
                               )

        # set default selection to 0

        if len(self.testRunFiles) > 0:
            # if testrun file not in Combo box of TestRunFiles
            index = self.testRunComboBox_4.findText(
                                 self.testRunFile,
                                 QtCore.Qt.MatchFixedString
                                 )
            if self.testRunFile not in self.testRunFiles:
                self.testRunComboBox_4.setCurrentIndex(0)
                self.testRunFile = self.testRunComboBox_4.currentText()
            else:
                self.testRunComboBox_4.setCurrentIndex(index)
            self.statusMessage("testrun file: {}".format(self.testRunFile))
            # Activate the Execute Button
            self.executePushButton_4.setEnabled(True)

        if len(self.configFiles) > 0:
            # if config file not in list of ConfigListFiles
            index = self.settingComboBox_4.findText(
                              self.configFile,
                              QtCore.Qt.MatchFixedString
                              )
            if self.configFile not in self.configFiles:
                self.settingComboBox_4.setCurrentIndex(0)
                # Activate the Settings Detail Button
                self.configFile = self.settingComboBox_4.currentText()
            else:
                self.settingComboBox_4.setCurrentIndex(index)
            self.statusMessage("value of self.configfile {}".format(
                               self.configFile))
            self.settingsPushButton_4.setEnabled(True)
            self.updateSettings()

    def setupBasePath(self, dirPath=""):
        """ Setup Base path of Execution as per directory Path"""

        if not dirPath:
            # Set up base path to Baangt directory
            # Based on current File path ../../../
            dirPath = os.path.dirname(os.path.dirname(
                          os.path.dirname(os.path.dirname(__file__))
                          ))
            if not dirPath:
                dirPath = os.path.abspath(os.curdir)
            self.pathLineEdit_4.insert(dirPath)
        else:
            self.pathLineEdit_4.setText(dirPath)
        self.directory = dirPath
        self.getSettingsAndTestFilesInDirectory(dirPath)
        self.statusMessage("Current Path: {} ".format(dirPath), 2000)

    @pyqtSlot()
    def quitApplication(self):
        """ This function will close the UI """
        buttonReply = QtWidgets.QMessageBox.question(
                           self.centralwidget,
                           "Close Confirmation ",
                           "Are you sure to quit?",
                           QtWidgets.QMessageBox.Yes |
                           QtWidgets.QMessageBox.No,
                           QtWidgets.QMessageBox.No
                           )
        if buttonReply == QtWidgets.QMessageBox.Yes:
            QtWidgets.QApplication.exit()

    @pyqtSlot()
    def mainPageView(self):
        """ This function will redirect to main page """
        self.stackedWidget.setCurrentIndex(0)

    @pyqtSlot()
    def updateRunFile(self):
        """ this file will update the testRunFile selection
        """
        self.testRunFile = os.path.join(self.directory,
                                  self.testRunComboBox_4.currentText())
        self.statusMessage("Test Run Changed to: {}".format(self.testRunFile))
        self.saveInteractiveGuiConfig()

    @pyqtSlot()
    def updateSettings(self):
        """ Update the settings Variable with content in fileName"""
        # Try to get full path
        # write changes to ini file
        self.configFile = os.path.join(self.directory,
                                 self.settingComboBox_4.currentText())

        self.saveInteractiveGuiConfig()
        self.statusMessage("Settings changed to: {}".format(self.configFile))
        self.readContentofGlobals()

    def executeButtonClicked(self):
        self.__result_file = ""
        self.__log_file = ""
        self.__log_file_monitor = DownloadFolderMonitoring(self.managedPaths.getLogfilePath())
        if self.__execute_button_state == "idle":
            self.runTestRun()
        elif self.__execute_button_state == "running":
            self.stopButtonPressed()

    @pyqtSlot()
    def runTestRun(self):
        if not self.configFile:
            self.statusMessage("No Config File", 2000)
            return
        if not self.testRunFile:
            self.statusMessage("No test Run File selected", 2000)
            return

        self.__execute_button_state = "running"
        self.stopIcon = QtGui.QIcon(":/baangt/stopicon")
        self.executePushButton_4.setIcon(self.stopIcon)
        self.executePushButton_4.setIconSize(QtCore.QSize(28, 20))
        self.executePushButton_4.setStyleSheet("color: rgb(255, 255, 255); background-color: rgb(255, 75, 50);")
        self.clear_logs_and_stats()

        runCmd = self._getRunCommand()

        # show status in status bar
        self.statusMessage("Executing.....", 4000)

        if self.configContents.get("TX.DEBUG", False) == "True":
            from baangt.base.TestRun.TestRun import TestRun

            lUUID = uuid4()
            self.lTestRun = ""
            self.lTestRun = TestRun(f"{Path(self.directory).joinpath(self.testRunFile)}",
                 globalSettingsFileNameAndPath=f'{Path(self.directory).joinpath(self.tempConfigFile)}', uuid=lUUID)
            self.processFinished(debug=True)

        else:
            logger.info(f"Running command: {runCmd}")
            self.run_process = QtCore.QProcess()
            self.run_process.setProcessChannelMode(QtCore.QProcess.MergedChannels)
            self.run_process.readyReadStandardOutput.connect(
                lambda: self.process_stdout(self.run_process.readAllStandardOutput()))
            self.run_process.readyReadStandardError.connect(
                lambda: self.logTextBox.appendPlainText(
                    str(self.run_process.readAllStandardError().data().decode('iso-8859-1'))))
            self.run_process.finished.connect(self.processFinished)
            self.run_process.start(runCmd)
            self.statusbar.showMessage("Running.....")

    @pyqtSlot()
    def stopButtonPressed(self):
        self.run_process.kill()

    @pyqtSlot()
    def processFinished(self, debug=False):
        buttonReply = QtWidgets.QMessageBox.question(
            self.centralwidget,
            "Baangt Interactive Starter ",
            "Test Run finished !!",
            QtWidgets.QMessageBox.Ok,
            QtWidgets.QMessageBox.Ok
        )
        if debug:
            self.__result_file = self.lTestRun.results.fileName
        self.statusMessage(f"Completed ", 3000)
        self.executeIcon = QtGui.QIcon(":/baangt/executeicon")
        self.executePushButton_4.setIcon(self.executeIcon)
        self.executePushButton_4.setIconSize(QtCore.QSize(28, 20))
        self.executePushButton_4.setStyleSheet("color: rgb(255, 255, 255); background-color: rgb(138, 226, 52);")
        self.__execute_button_state = "idle"
        try:
            self.__log_file = self.__log_file_monitor.getNewFiles()[0][0]
        except:
            pass

        # Remove temporary Configfile, that was created only for this run:
        try:
            os.remove(Path(self.directory).joinpath(self.tempConfigFile))
        except Exception as e:
            logger.warning(f"Tried to remove temporary file but seems to be not there: "
                           f"{self.directory}/{self.tempConfigFile}")
        if self.__open_files:
            self.openResultFile()
            self.openLogFile()

    def process_stdout(self, obj):
        text = str(obj.data().decode('iso-8859-1'))
        if "ExportResults _ __init__ : Export-Sheet for results: " in text:
            lis = text[text.index("results: "):].split(" ")
            result_file = lis[1].strip()
            while len(result_file)>4:
                if result_file[-5:] != ".xlsx":
                    result_file = result_file[:-1]
                else:
                    break
            if result_file[-5:] == ".xlsx":
                self.__result_file =result_file
        if "||Statistic:" in text:
            lis = text.split('||')
            stat = lis[1][10:]
            stat_lis = stat.split('\n')
            log = ''.join([lis[0], lis[2]])
            for x in range(9):
                self.statisticTable.setItem(0, x, QtWidgets.QTableWidgetItem(stat_lis[x].split(': ')[1]))
                self.statisticTable.item(0,x).setTextAlignment(QtCore.Qt.AlignCenter)
                self.statisticTable.item(0, x).setFlags(
                    self.statisticTable.item(0, x).flags() ^ QtCore.Qt.ItemIsSelectable)
                self.statisticTable.item(0, x).setFlags(
                    self.statisticTable.item(0, x).flags() ^ QtCore.Qt.ItemIsEditable)
                if x == 3:
                    self.statisticTable.item(0, 3).setBackground(QtGui.QBrush(QtCore.Qt.green))
                elif x == 4:
                    self.statisticTable.item(0, 4).setBackground(QtGui.QBrush(QtCore.Qt.red))
                else:
                    self.statisticTable.item(0, x).setBackground(QtGui.QBrush(QtCore.Qt.white))
            log.replace('\n', '')
            if not log.strip() == "":
                self.logTextBox.appendPlainText(log.strip())
        else:
            text = text.replace('\n', '')
            if not text.strip() == "":
                self.logTextBox.appendPlainText(text.strip())

    def _getRunCommand(self):
        """
        If bundled (e.g. in pyinstaller),
        then the executable is already sys.executable,
        otherwise we need to concatenate executable and
        Script-Name before we can start
        a subprocess.

        @return: Full path and filename to call Subprocess
        """
        lStart = sys.executable
        if "python" in sys.executable.lower():
            if len(Path(sys.argv[0]).parents) > 1:
                # This is a system where the path the the script is
                # given in sys.argv[0]
                lStart = f'"{lStart}"' + f' "{sys.argv[0]}"'
            else:
                # this is a system where we need to join os.getcwd()
                # and sys.argv[0] because the path is not given in sys.argv[0]
                lStart = f'"{lStart}"' + f' "{Path(os.getcwd()).joinpath(sys.argv[0])}"'
        else:
            lStart = f'"{lStart}"'

        self.__makeTempConfigFile()

        return f'{lStart} ' \
               f'--run="{Path(self.directory).joinpath(self.testRunFile)}" ' \
               f'--globals="{Path(self.directory).joinpath(self.tempConfigFile)}" --gui True'


    def __makeTempConfigFile(self):
        """
         Add parameters to the Config-File for this Testrun and
         save the file under a temporary name
        """
        self.configContents[GC.PATH_ROOT] = self.directory
        self.configContents[GC.PATH_SCREENSHOTS] = str(Path(
                                  self.directory).joinpath("Screenshots"))
        self.configContents[GC.PATH_EXPORT] = str(Path(
                                  self.directory).joinpath("1testoutput"))
        self.configContents[GC.PATH_IMPORT] = str(Path(
                                  self.directory).joinpath("0testdateninput"))
        self.tempConfigFile = MainWindow.__makeRandomFileName()
        self.saveContentsOfConfigFile(self.tempConfigFile)

    def saveContentsOfConfigFile(self, lFileName = None):
        if not lFileName:
            lFileName = self.configFile

        with open(str(Path(self.directory).joinpath(lFileName)), 'w') as outfile:
            json.dump(self.configContents, outfile, indent=4)


    @staticmethod
    def __makeRandomFileName():
        return "globals_" + utils.datetime_return() + ".json"

    @pyqtSlot()
    def settingView(self):
        """
        View settings Below Main Windows
        """
        self.statusMessage("Settings Page Opened")
        self.stackedWidget.setCurrentIndex(1)
        self.readConfigFile()
        self.drawSetting()

    @pyqtSlot()
    def browsePathSlot(self):
        """ Browse Folder Containing *.xlsx file for execution. And
           globals.json file for Test specific settings
        """
        # get path from pathLineEdit
        basepath = self.pathLineEdit_4.text()
        if not basepath:
            basepath = "./"
        options = QtWidgets.QFileDialog.Options()
        options |= QtWidgets.QFileDialog.DontUseNativeDialog
        dirName = QtWidgets.QFileDialog.getExistingDirectory(
                         None,
                         "Select Directory ",
                         basepath,
                         options=options
                         )
        if dirName:
            # self.pathLineEdit.insert(dirName)
            self.setupBasePath(dirName)

    # Settings Page
    # All action and function related to Setting page is below
    def readConfigFile(self):
        """ Read the configFile and update the configInstance """
        if self.configFile:
            # Compute full path
            fullpath = os.path.join(self.directory, self.configFile)
            if os.path.isfile(fullpath):
                self.configInstance = GlobalSettings.getInstance()
                self.configInstance.addValue(fullpath)

    @QtCore.pyqtSlot()
    def addMore(self):
        """ This function will popup a dialog box,
        User input the keword and a new row is added
        to Form Layout
        """
        # get total no of rows, it will be index for new row
        count = self.formLayout.rowCount()
        all_keys = self.configInstance.globalconfig.items()
        # get keys values from formLayout
        formlayoutItems = self.parseFormLayout()

        # convert to dict, to fix unhashable type: dict error
        all_keys = dict(all_keys)
        # unused keys : globalconfig keys - formlayout keys
        keys = [d for d in all_keys.keys() if d not in formlayoutItems]

        # Now prepare key and displayText pair
        displayTextPairs = [
                         (k, v['displayText'])
                         for k, v in all_keys.items()
                         if k in keys]

        # to store keys for displayText list
        shownkeys = [p[0] for p in displayTextPairs]
        # to store displayTextList
        shownvalues = [p[1] for p in displayTextPairs]

        item, okPressed = QtWidgets.QInputDialog.getItem(
                               None,
                               "New Parameter ",
                               "Parameter Name",
                               shownvalues,
                               0,
                               True
                               )
        if item and okPressed:
            if item in shownvalues:
                # get keys using index of item selection
                key = shownkeys[shownvalues.index(item)]
            else:
                # this is the key still not added in globalSettings
                key = item
            value = self.configInstance.globalconfig.get(
                              key,
                              GlobalSettings.transformToDict(key, ""))
            self.addNewRow(count, key, value)

    @QtCore.pyqtSlot()
    def deleteLast(self):
        """ This function when call delete last row
        from the form layout
        """
        # get the index of last row, equals totalrow minus one
        count = self.formLayout.rowCount()

        # delete the count - 1 th row
        if count > 0:
            self.formLayout.removeRow(count - 1)

    @QtCore.pyqtSlot()
    def saveAsNewFile(self):
        """ This will ask new file to save data """
        # save recent changes to config
        self.saveValue()
        # ask for fileName
        options = QtWidgets.QFileDialog.Options()
        options |= QtWidgets.QFileDialog.DontUseNativeDialog
        filename = QtWidgets.QFileDialog.getSaveFileName(
                                None,
                                "Save Global Setting  File",
                                self.directory,
                                "JsonFile (*.json)",
                                "",
                                options=options
                                )
        newFile = filename[0]
        if newFile:
            if not os.path.basename(newFile).endswith(".json"):
                newFile = os.path.join(
                             os.path.dirname(newFile),
                             os.path.basename(newFile) + ".json"
                             )

            data = {}
            for key, value in self.configInstance.config.items():
                data[key] = value

            with open(newFile, 'w') as f:
                json.dump(data, f, indent=4)

    @QtCore.pyqtSlot()
    def saveToFile(self):
        """ Save the content to file"""
        # call saveFile before saving to File
        self.saveValue()
        data = {}
        for key, value in self.configInstance.config.items():
            data[key] = value

        if not self.configFile:
            # Open Dialog box to save file
            options = QtWidgets.QFileDialog.Options()
            options |= QtWidgets.QFileDialog.DontUseNativeDialog
            filename = QtWidgets.QFileDialog.getSaveFileName(
                                    None,
                                    "Save Global Setting  File",
                                    self.directory,
                                    "JsonFile (*.json)",
                                    "",
                                    options=options
                                    )
            self.configFile = filename[0]
        if self.configFile:
            if not os.path.basename(self.configFile).endswith(".json"):
                self.configFile = os.path.join(
                             os.path.dirname(self.configFile),
                             os.path.basename(self.configFile) + ".json"
                             )

            fullpath = self.configFile
            if not os.path.isabs(self.configFile):
                if self.directory:
                    fullpath = os.path.join(self.directory, self.configFile)
                else:
                    self.directory = os.getcwd()#.managedPaths.derivePathForOSAndInstallationOption()
                    fullpath = os.path.join(self.directory, self.configFile)

            with open(fullpath, 'w') as f:
                json.dump(data, f, indent=4)

        self.drawSetting()
        self.readContentofGlobals()
        self.mainPageView()

    def parseFormLayout(self):
        """ This function will parse form layout
        and return dictionary items
        """
        data = {}
        count = self.formLayout.rowCount()
        for d in range(count):

            # item = self.formLayout.takeRow(0)
            labelItem = self.formLayout.itemAt(
                                   d,
                                   QtWidgets.QFormLayout.LabelRole
                                   )

            fieldItem = self.formLayout.itemAt(
                                   d,
                                   QtWidgets.QFormLayout.FieldRole
                                   )
            # print(labelItem)
            # print(fieldItem)
            key = ""
            value = ""
            if isinstance(labelItem, QtWidgets.QWidgetItem):
                lablename = labelItem.widget()
                key = lablename.objectName()
            if isinstance(fieldItem, QtWidgets.QWidgetItem):
                fieldname = fieldItem.widget()
                if isinstance(fieldname, QtWidgets.QCheckBox):
                    # get checked status
                    value = str(fieldname.isChecked())
                elif isinstance(fieldname, QtWidgets.QComboBox):
                    # get current Text
                    value = fieldname.currentText()
                elif isinstance(fieldname, QtWidgets.QLineEdit):
                    value = fieldname.text()
            if key:
                data[key] = value

        return data

    @QtCore.pyqtSlot()
    def saveValue(self):
        """ This simple function call parseFormlayout to get
        dictionary data and update the config value
        """
        # update the data to config instance
        data = self.parseFormLayout()

        if self.configInstance:
            self.configInstance.updateValue(data)
            # print(self.configInstance.config)
        else:
            print("No config instance ")
            self.configInstance = GlobalSettings.getInstance()
            self.configInstance.updateValue(data)

    def drawSetting(self):
        """ This will draw Setting based on data in configInstance
        """
        # We will use filtered dict key only
        # settings = self.configInstance.filterIniKey()
        # Remove all existing data to print
        n_rows = self.formLayout.rowCount()

        if n_rows > 0:
            # Delete all existing rows
            # print("Number of rows", n_rows)
            for d in range(n_rows):
                self.formLayout.removeRow(0)
        self.formLayout.update()

        # update the groupbox headlines
        if self.configFile:
            settingFile = self.configFile
            # compute full path
            fullpath = os.path.join(self.directory, self.configFile)
            if os.path.isfile(fullpath):
                settingFile = fullpath
        else:
            settingFile = "globalSetting.json"
        _translate = QtCore.QCoreApplication.translate
        self.groupBox.setTitle(
                   _translate(
                       "Form",
                       "Settings in {}".format(
                          os.path.basename(settingFile)
                       )))
        # prepare settings here
        settings = {}
        for key, value in self.configInstance.config.items():
            if key in self.configInstance.globalconfig:
                settings[key] = self.configInstance.globalconfig[key]
                settings[key]['default'] = value
            else:
                settings[key] = GlobalSettings.transformToDict(key, value)

        # settings = self.configInstance.config
        count = 0
        for key, value in sorted(
                         settings.items(),
                         key=lambda x: x[1]['type']
                         ):
            self.addNewRow(count, key, value)
            count += 1

    def addNewRow(self, count, key, value):
        """ This function will add new row at
        count number with given key value pair
        in formlayout
        """
        _translate = QtCore.QCoreApplication.translate
        if value['type'] == 'bool':
            # create check box
            self.checkBox1Label = QtWidgets.QLabel(
                                 self.scrollAreaWidgetContents
                                  )
            self.checkBox1Label.setObjectName(key)
            self.checkBox1Label.setToolTip(value['hint'])
            self.checkBox1Label.setText(
                           _translate("Form", value['displayText']))
            self.checkBox1CheckBox = QtWidgets.QCheckBox(
                            self.scrollAreaWidgetContents)
            # self.checkBox1CheckBox.setStyleSheet(
            #                "color: rgb(46, 52, 54);")
            if isinstance(value['default'], bool):
                # its bool type
                self.checkBox1CheckBox.setChecked(value['default'])

            elif isinstance(value['default'], str):
                # its string type
                if value['default'].lower() == "true":
                    self.checkBox1CheckBox.setChecked(True)
                else:
                    self.checkBox1CheckBox.setChecked(False)
            else:
                # default
                self.checkBox1CheckBox.setChecked(False)
            self.formLayout.setWidget(
                              count,
                              QtWidgets.QFormLayout.LabelRole,
                              self.checkBox1Label)
            self.formLayout.setWidget(
                              count,
                              QtWidgets.QFormLayout.FieldRole,
                              self.checkBox1CheckBox)

        elif value['type'] == 'text':
            self.lineEdit1Label = QtWidgets.QLabel(
                              self.scrollAreaWidgetContents)
            self.lineEdit1Label.setToolTip(
                               _translate("Form", value['hint']))
            self.lineEdit1Label.setObjectName(key)
            self.lineEdit1Label.setText(
                               _translate("Form", value['displayText'])
                               )
            self.lineEdit1LineEdit = QtWidgets.QLineEdit(
                              self.scrollAreaWidgetContents)
            sizePolicy = QtWidgets.QSizePolicy(
                        QtWidgets.QSizePolicy.MinimumExpanding,
                        QtWidgets.QSizePolicy.Fixed)
            sizePolicy.setHorizontalStretch(0)
            sizePolicy.setVerticalStretch(0)
            sizePolicy.setHeightForWidth(
                    self.lineEdit1LineEdit.sizePolicy().hasHeightForWidth())
            self.lineEdit1LineEdit.setSizePolicy(sizePolicy)
            self.lineEdit1LineEdit.setMinimumSize(QtCore.QSize(250, 0))

            # self.lineEdit1LineEdit.setStyleSheet(
            #                 "background-color: rgb(255, 255, 255);\n"
            #                 "color: rgb(46, 52, 54);")
            self.lineEdit1LineEdit.setText(
                             _translate("Form", value['default']))
            self.formLayout.setWidget(
                             count,
                             QtWidgets.QFormLayout.LabelRole,
                             self.lineEdit1Label)
            self.formLayout.setWidget(
                             count,
                             QtWidgets.QFormLayout.FieldRole,
                             self.lineEdit1LineEdit)

        elif value['type'] == 'select':
            self.comboBox1Label = QtWidgets.QLabel(
                             self.scrollAreaWidgetContents)
            self.comboBox1Label.setObjectName(key)
            self.comboBox1Label.setToolTip(value['hint'])
            self.comboBox1Label.setText(
                             _translate("Form", value['displayText']))
            self.formLayout.setWidget(
                              count,
                              QtWidgets.QFormLayout.LabelRole,
                              self.comboBox1Label)
            self.comboBox1ComboBox = QtWidgets.QComboBox(
                              self.scrollAreaWidgetContents)
            # self.comboBox1ComboBox.setStyleSheet(
            #                  "color: rgb(46, 52, 54):\n"
            #                  "background-color: rgb(255, 255, 255);")
            self.comboBox1ComboBox.addItems(value['options'])
            # set the Value
            self.comboBox1ComboBox.setCurrentIndex(
                            self.comboBox1ComboBox.findText(
                                value['default'],
                                QtCore.Qt.MatchFixedString
                                ))
            self.formLayout.setWidget(
                              count,
                              QtWidgets.QFormLayout.FieldRole,
                              self.comboBox1ComboBox)

    # Katalon Recorder Page
    #  All setting and Action for Katalon Page is Below
    @QtCore.pyqtSlot()
    def exitKatalon(self):
        """ this function will clear text from textIn and
        TextOut and exit
        """
        self.TextIn_2.clear()
        self.TextOut_2.clear()
        self.stackedWidget.setCurrentIndex(0)

    @QtCore.pyqtSlot()
    def saveTestCase(self):
        """ Use Existing ImportKatalonRecorder.saveTestCase internally to
        save test cast to XLSX
        """
        options = QtWidgets.QFileDialog.Options()
        options |= QtWidgets.QFileDialog.DontUseNativeDialog

        filename = QtWidgets.QFileDialog.getSaveFileName(
                                None,
                                "Save Test Case File",
                                self.katalonRecorder.directory,
                                "",
                                "",
                                options=options
                                )
        # resulted filename is in tuple, (fullpath, All Files(*))
        if filename[0]:
            self.katalonRecorder.directory = os.path.dirname(filename[0])
            self.katalonRecorder.fileNameExport = os.path.basename(filename[0])
            # save File
            self.katalonRecorder.saveTestCase()

            # Change the testrun file to newone and change directory
            self.testRunFile = self.katalonRecorder.fileNameExport
            self.setupBasePath(self.katalonRecorder.directory)
            self.exitKatalon()


    @QtCore.pyqtSlot()
    def importClipboard(self):
        """Extend: katalonRecorder.importClipboard internally """
        # ignore last line as this will result unexpected
        # Index out of range error
        self.clipboardText = self.TextIn_2.toPlainText()
        self.katalonRecorder.clipboardText = "\n".join([
                              text for text in self.clipboardText.split("\n")
                              if len(text.split("|")) > 2
                              ])
        self.katalonRecorder.importClipboard()
        self.TextOut_2.setPlainText(self.katalonRecorder.outputText)

    @QtCore.pyqtSlot()
    def copyFromClipboard(self):
        """ Call ImportKatalonRecorder.importClipboard internally """
        self.TextIn_2.setPlainText(pyperclip.paste())
        self.importClipboard()

    @QtCore.pyqtSlot()
    def openResultFile(self):
        """ Uses Files Open class to open Result file """
        try:
            if self.__result_file != "":
                filePathName = self.__result_file
                fileName = os.path.basename(filePathName)
                self.statusMessage(f"Opening file {fileName}", 3000)
                FilesOpen.openResultFile(filePathName)
            else:
                self.statusMessage("No file found!", 3000)
        except:
            self.statusMessage("No file found!", 3000)

    @QtCore.pyqtSlot()
    def openLogFile(self):
        """ Uses Files Open class to open Log file """
        try:
            if self.__result_file != "":
                wb = xlrd.open_workbook(self.__result_file)
                book = wb.sheet_by_name("Summary")
                filePathName = book.row(7)[1].value.strip()
                fileName = os.path.basename(filePathName)
                self.statusMessage(f"Opening file {fileName}", 3000)
                FilesOpen.openResultFile(filePathName)
            else:
                if self.__log_file != "":
                    filePathName = self.__log_file
                    fileName = os.path.basename(filePathName)
                    self.statusMessage(f"Opening file {fileName}", 3000)
                    FilesOpen.openResultFile(filePathName)
                else:
                    self.statusMessage("No file found!", 3000)
        except:
            self.statusMessage("No file found!", 3000)

    @QtCore.pyqtSlot()
    def openTestFile(self):
        """ Uses Files Open class to open Log file """
        try:
            filePathName = f"{Path(self.directory).joinpath(self.testRunFile)}"
            fileName = os.path.basename(filePathName)
            self.statusbar.showMessage(f"Opening file {fileName}")
            FilesOpen.openResultFile(filePathName)
        except:
            self.statusMessage("No file found!", 3000)

    @pyqtSlot()
    def clear_logs_and_stats(self):
        for x in range(9):
            self.statisticTable.setItem(0, x, QtWidgets.QTableWidgetItem())
            self.statisticTable.item(0,x).setBackground(QtGui.QBrush(QtCore.Qt.white))
        self.logTextBox.clear()
        QtCore.QCoreApplication.processEvents()

    @pyqtSlot()
    def show_hide_logs(self):
        if self.logSwitch.isChecked():
            self.logTextBox.show()
            self.__log_state = 1
        else:
            self.logTextBox.hide()
            self.__log_state = 0
        self.saveInteractiveGuiConfig()

    def change_openFiles_state(self):
        if self.openFilesSwitch.isChecked():
            self.__open_files = 1
        else:
            self.__open_files = 0
        self.saveInteractiveGuiConfig()


# Controller
class MainController:
    def __init__(self):
        self.widget = QtWidgets.QWidget()
        self.window = QtWidgets.QMainWindow()
        self.main = MainWindow()

    def show_main(self):
        self.main = MainWindow()
        self.main.setupUi(self.window)
        self.window.show()


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    controller = MainController()
    controller.show_main()
    sys.exit(app.exec_())



from dataclasses import dataclass

import sys


class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


@dataclass
class Statistic(metaclass=Singleton):
    testcases_total: int = 0
    testcases_executed: int = 0
    testcases_remaining: int = 0
    testcases_success: int = 0
    testcases_failed: int = 0
    testcases_paused: int = 0
    testcase_sequence_executed: int = 0
    teststep_executed: int = 0
    teststep_sequence_executed: int = 0
    gui: bool = False

    def __str__(self):
        string = [f"{key.split('_')[0]+' '+key.split('_')[1]}: {self.__dict__[key]}" for key in self.__dict__ if "_" in key]
        return "\n".join(string).upper()

    def total_testcases(self, number):
        self.testcases_total = number
        self.testcases_remaining = number
        if self.gui:
            sys.stdout.write("||Statistic:"+self.__str__()+"||")
            sys.stdout.flush()

    def update_all(self, success, error, waiting):
        self.testcases_executed = success+error
        self.testcases_remaining = self.testcases_total - self.testcases_executed
        self.testcases_paused = waiting
        self.testcases_success = success
        self.testcases_failed = error
        if self.gui:
            sys.stdout.write("||Statistic:"+self.__str__()+"||")
            sys.stdout.flush()

    def update_teststep(self):
        self.teststep_executed += 1
        if self.gui:
            sys.stdout.write("||Statistic:"+self.__str__()+"||")
            sys.stdout.flush()

    def update_testcase_sequence(self):
        self.testcase_sequence_executed += 1
        if self.gui:
            sys.stdout.write("||Statistic:"+self.__str__()+"||")
            sys.stdout.flush()

    def update_teststep_sequence(self):
        self.teststep_sequence_executed += 1
        if self.gui:
            sys.stdout.write("||Statistic:"+self.__str__()+"||")
            sys.stdout.flush()






import baangt.base.GlobalConstants as GC
from baangt.base.Timing.Timing import Timing
from baangt.base.BrowserHandling.BrowserHandling import BrowserDriver
from baangt.base.ApiHandling import ApiHandling
import sys
from pkg_resources import parse_version
import logging
from baangt.TestSteps.Exceptions import baangtTestStepException
from baangt.TestSteps.AddressCreation import AddressCreate
from baangt.base.PDFCompare import PDFCompare, PDFCompareDetails
from baangt.base.Faker import Faker as baangtFaker
from baangt.base.Utils import utils
from baangt.base.RuntimeStatistics import Statistic

logger = logging.getLogger("pyC")


class TestStepMaster:
    def __init__(self, executeDirect=True, **kwargs):
        self.anchor = None
        self.anchorLocator = None
        self.anchorLocatorType = None
        self.testCaseStatus = None
        self.ifActive = False
        self.ifIsTrue = True
        self.baangtFaker = None
        self.statistics = Statistic()
        self.kwargs = kwargs
        self.testRunInstance = kwargs.get(GC.KWARGS_TESTRUNINSTANCE)
        self.testcaseDataDict = kwargs.get(GC.KWARGS_DATA)
        self.timing: Timing = kwargs.get(GC.KWARGS_TIMING)
        self.timingName = self.timing.takeTime(self.__class__.__name__, forceNew=True)
        self.browserSession: BrowserDriver = kwargs.get(GC.KWARGS_BROWSER)
        self.apiSession: ApiHandling = kwargs.get(GC.KWARGS_API_SESSION)

        self.testStepNumber = kwargs.get(GC.STRUCTURE_TESTSTEP)  # Set in TestRunData by TestCaseMaster
        self.testRunUtil = self.testRunInstance.testRunUtils

        # Get the release from testrun-customizing if set: (will influence, whether or not elements will be executed)
        self.globalRelease = self.testRunInstance.globalSettings.get("Release", "")

        # check, if this TestStep has additional Parameters and if so, execute
        lSequence = self.testRunUtil.getSequenceByNumber(testRunName=self.testRunInstance.testRunName,
                                                         sequence=kwargs.get(GC.STRUCTURE_TESTCASESEQUENCE))
        self.testCase = self.testRunUtil.getTestCaseByNumber(lSequence, kwargs.get(GC.STRUCTURE_TESTCASE))
        self.testStep = self.testRunUtil.getTestStepByNumber(self.testCase, self.testStepNumber)

        if self.testStep and len(self.testStep) > 1:
            if not isinstance(self.testStep[1], str) and executeDirect:
                # This TestStepMaster-Instance should actually do something - activitites are described
                # in the TestExecutionSteps.
                # Otherwise there's only a classname in TestStep[0]
                self.executeDirect(self.testStep[1][GC.STRUCTURE_TESTSTEPEXECUTION])

                # Teardown makes only sense, when we actually executed something directly in here
                # Otherwise (if it was 1 or 2 Tab-stops more to the left) we'd take execution time without
                # having done anything
                self.teardown()

        self.statistics.update_teststep_sequence()

    def executeDirect(self, executionCommands):
        """
        Executes a sequence of Commands. Will be subclassed in other modules.
        :param executionCommands:
        :return:
        """
        for index, (key, command) in enumerate(executionCommands.items()):
            self.executeDirectSingle(index, command)
            # self.statistics.update_teststep()  --> Moved to BrowserHandling into the activities themselves,
            # so that we can also count for externally (subclassed) activitis


    def executeDirectSingle(self, commandNumber, command):
        """
        This will execute a single instruction
        """

        # when we have an IF-condition and it's condition was not TRUE, then skip whatever comes here until we
        # reach Endif
        if not self.ifIsTrue and command["Activity"] != "ENDIF":
            return
        lActivity = command["Activity"].upper()
        if lActivity == "COMMENT":
            return  # Comment's are ignored

        lLocatorType = command["LocatorType"].upper()
        lLocator = self.replaceVariables(command["Locator"])

        if lLocator and not lLocatorType:  # If locatorType is empty, default it to XPATH
            lLocatorType = 'XPATH'

        xpath, css, id = self.__setLocator(lLocatorType, lLocator)

        if self.anchor and xpath:
            if xpath[0:3] == '///':         # Xpath doesn't want to use Anchor
                xpath = xpath[1:]
                logger.debug(f"Anchor active, but escaped. Using pure xpath: {xpath}")
            else:
                logger.debug(f"Anchor active. combining {self.anchorLocator} with {xpath}")
                xpath = self.anchorLocator + xpath

        lValue = str(command["Value"])
        lValue2 = str(command["Value2"])
        lComparison = command["Comparison"]
        lOptional = utils.anything2Boolean(command["Optional"])

        # check release line
        lRelease = command["Release"]

        # Timeout defaults to 20 seconds, if not set otherwise.
        lTimeout = TestStepMaster.__setTimeout(command["Timeout"])

        lTimingString = f"TS {commandNumber} {lActivity.lower()}"
        self.timing.takeTime(lTimingString)
        logger.info(
            f"Executing TestStepDetail {commandNumber} with parameters: act={lActivity}, lType={lLocatorType}, loc={lLocator}, "
            f"Val1={lValue}, comp={lComparison}, Val2={lValue2}, Optional={lOptional}, timeout={lTimeout}")

        lValue, lValue2 = self.replaceAllVariables(lValue, lValue2)

        if not TestStepMaster.ifQualifyForExecution(self.globalRelease, lRelease):
            logger.debug(f"we skipped this line due to {lRelease} disqualifies according to {self.globalRelease} ")
            return
        if lActivity == "GOTOURL":
            self.browserSession.goToUrl(lValue)
        elif lActivity == "SETTEXT":
            self.browserSession.findByAndSetText(xpath=xpath, css=css, id=id, value=lValue, timeout=lTimeout)
        elif lActivity == "SETTEXTIF":
            self.browserSession.findByAndSetTextIf(xpath=xpath, css=css, id=id, value=lValue, timeout=lTimeout,
                                                   optional=lOptional)
        elif lActivity == "FORCETEXT":
            self.browserSession.findByAndForceText(xpath=xpath, css=css, id=id, value=lValue, timeout=lTimeout)
        elif lActivity == "SETANCHOR":
            if not lLocator:
                self.anchor = None
                self.anchorLocator = None
                self.anchorLocatorType = None
            else:
                found = self.browserSession.findBy(xpath=xpath, css=css, id=id, timeout=lTimeout)
                if found:
                    self.anchor = self.browserSession.element
                    self.anchorLocator = lLocator
                    self.anchorLocatorType = lLocatorType
                    logger.debug(f"Anchor set: {lLocatorType} {lLocator}")
                else:
                    logger.error(f"Anchor should be set, but can't be found in the current page: {lLocatorType}, {lLocator}")
                    raise ValueError(f"Anchor should be set, but can't be found in the current page: {lLocatorType}, {lLocator}")
        elif lActivity == 'HANDLEIFRAME':
            self.browserSession.handleIframe(lLocator)
        elif lActivity == 'SWITCHWINDOW':
            lWindow = lValue
            if lWindow.isnumeric():
                lWindow = int(lWindow)
            self.browserSession.handleWindow(windowNumber=lWindow, timeout=lTimeout)
        elif lActivity == "CLICK":
            self.browserSession.findByAndClick(xpath=xpath, css=css, id=id, timeout=lTimeout)
        elif lActivity == "CLICKIF":
            self.browserSession.findByAndClickIf(xpath=xpath, css=css, id=id, timeout=lTimeout, value=lValue)
        elif lActivity == "PAUSE":
            self.browserSession.sleep(seconds=float(lValue))
        elif lActivity == "IF":
            if self.ifActive:
                raise BaseException("No nested IFs at this point, sorry...")
            self.ifActive = True
            # Originally we had only Comparisons. Now we also want to check for existance of Field
            if not lValue and lLocatorType and lLocator:
                lValue = self.browserSession.findBy(xpath=xpath, css=css, id=id, optional=lOptional,
                                                    timeout=lTimeout)

            self.__doComparisons(lComparison=lComparison, value1=lValue, value2=lValue2)
            logger.debug(f"IF-condition {lValue} {lComparison} {lValue2} evaluated to: {self.ifIsTrue} ")
        elif lActivity == "ENDIF":
            if not self.ifActive:
                raise BaseException("ENDIF without IF")
            self.ifActive = False
            self.ifIsTrue = True
        elif lActivity == 'GOBACK':
            self.browserSession.goBack()
        elif lActivity == 'APIURL':
            self.apiSession.setBaseURL(lValue)
        elif lActivity == 'ENDPOINT':
            self.apiSession.setEndPoint(lValue)
        elif lActivity == 'POST':
            self.apiSession.postURL(content=lValue)
        elif lActivity == 'GET':
            self.apiSession.getURL()
        elif lActivity == 'HEADER':
            self.apiSession.setHeaders(setHeaderData=lValue)
        elif lActivity == 'SAVE':
            self.doSaveData(lValue, lValue2, lLocatorType, lLocator)
        elif lActivity == 'CLEAR':
            # Clear a variable:
            if self.testcaseDataDict.get(lValue):
                del self.testcaseDataDict[lValue]
        elif lActivity == 'SAVETO':
            # In this case, we need to parse the real field, not the representation of the replaced field value
            self.doSaveData(command['Value'], lValue2, lLocatorType, lLocator)
        elif lActivity == 'SUBMIT':
            self.browserSession.submit()
        elif lActivity == "ADDRESS_CREATE":
            # Create Address with option lValue and lValue2
            AddressCreate(lValue, lValue2)
            # Update testcaseDataDict with addressDict returned from
            AddressCreate.returnAddress()
            self.testcaseDataDict.update(AddressCreate.returnAddress())
        elif lActivity == 'ASSERT':
            value_found = self.browserSession.findByAndWaitForValue(xpath=xpath, css=css, id=id, optional=lOptional,
                                                                    timeout=lTimeout)
            if not self.__doComparisons(lComparison=lComparison, value1=value_found, value2=lValue):
                raise baangtTestStepException(f"Expected Value: {lValue}, Value found :{value_found} ")
        elif lActivity == 'IBAN':
            # Create Random IBAN. Value1 = Input-Parameter for IBAN-Function. Value2=Fieldname
            self.__getIBAN(lValue, lValue2)
        elif lActivity == 'PDFCOMPARE':
            self.doPDFComparison(lValue)
        elif lActivity == 'CHECKLINKS':
            self.checkLinks()
        elif lActivity == 'ALERTIF':
            self.browserSession.confirmAlertIfAny()
        else:
            raise BaseException(f"Unknown command in TestStep {lActivity}")

        self.timing.takeTime(lTimingString)

    def doPDFComparison(self, lValue, lFieldnameForResults="DOC_Compare"):
        lFiles = self.browserSession.findNewFiles()
        if len(lFiles) > 1:
            # fixme: Do something! There were more than 1 files since last check. Damn
            logger.critical(f"There were {len(lFiles)} files new since last check. Can't handle that. ")
            raise Exception
        elif len(lFiles) == 1:
            # Wonderful. Let's do the PDF-Comparison
            lPDFDataClass = PDFCompareDetails()
            lPDFDataClass.fileName = lFiles[0][0]
            lPDFDataClass.referenceID = lValue
            lDict = {"": lPDFDataClass}
            lPDFCompare = PDFCompare()
            lDict = lPDFCompare.compare_multiple(lDict)
            self.testcaseDataDict[lFieldnameForResults + "_Status"] = lDict[""].Status
            self.testcaseDataDict[lFieldnameForResults + "_Results"] = lDict[""].StatusText

    def replaceAllVariables(self, lValue, lValue2):
        # Replace variables from data file
        if len(lValue) > 0:
            lValue = self.replaceVariables(lValue)
        if len(lValue2) > 0:
            lValue2 = self.replaceVariables(lValue2)
        return lValue, lValue2

    def __getIBAN(self, lValue, lValue2):
        from baangt.base.IBAN import IBAN
        if not lValue2:
            logger.critical("IBAN-Method called without destination field name in column 'Value 2'")
            return

        lIBAN = IBAN()
        self.testcaseDataDict[lValue2] = lIBAN.getRandomIBAN()
        pass

    def checkLinks(self):
        """
        Will check all links on the current webpage

        Result will be written into "CheckedLinks" in TestDataDict

        If theres a returncode >= 400 in the list, we'll mark the testcase as failed
        """
        if self.testcaseDataDict.get("CheckedLinks"):
            self.testcaseDataDict["Result_CheckedLinks"] += self.browserSession.checkLinks()
        else:
            self.testcaseDataDict["Result_CheckedLinks"] = self.browserSession.checkLinks()

        for entry in self.testcaseDataDict["Result_CheckedLinks"]:
            if entry[0] > 400:
                self.testcaseDataDict[GC.TESTCASESTATUS] = GC.TESTCASESTATUS_ERROR
                self.testcaseDataDict[GC.TESTCASEERRORLOG] = self.testcaseDataDict.get(GC.TESTCASEERRORLOG, "") \
                                                             + "URL-Checker error"
                break

    @staticmethod
    def ifQualifyForExecution(version_global, version_line):
        """ This function will test version_global and version_line
            @return True or False
        """
        if not version_global.strip():
            # No value is defined in Release, return True
            return True
        if not version_line.strip():
            # we skipped this line
            return True

        # split the version line
        if not len(version_line.strip().split(" ")) == 2:
            logger.debug(f"Invalid release format {version_line} ")
            return True
        comp_operator, version = version_line.strip().split(" ")
        if comp_operator == "<":
            return parse_version(version_global) < parse_version(version)
        elif comp_operator == ">":
            return parse_version(version_global) > parse_version(version)
        elif comp_operator == ">=":
            return parse_version(version_global) >= parse_version(version)
        elif comp_operator == "<=":
            return parse_version(version_global) <= parse_version(version)
        elif comp_operator == "=" or comp_operator == "==":
            return parse_version(version_global) == parse_version(version)
        else:
            logger.debug(f"Global version {version_global}, line version {version_line} ")
            return False

    def doSaveData(self, toField, valueForField, lLocatorType, lLocator):
        """
        Save fields. Either from an existing DICT (usually in API-Mode) or from a Webelement (in Browser-Mode)

        :param toField:
        :param valueForField:
        :param lLocatorType:
        :param lLocator:
        :return: no return parameter. The implicit return is a value in a field.
        """
        if self.testCase[1][GC.KWARGS_TESTCASETYPE] == GC.KWARGS_BROWSER:
            xpath, css, id = TestStepMaster.__setLocator(lLocatorType, lLocator)
            self.testcaseDataDict[toField] = self.browserSession.findByAndWaitForValue(xpath=xpath, css=css, id=id)
            logger.debug(f"Saved {self.testcaseDataDict[toField]} to {toField}")
        elif self.testCase[1][GC.KWARGS_TESTCASETYPE] == GC.KWARGS_API_SESSION:
            self.testcaseDataDict[toField] = valueForField
        else:
            sys.exit("Testcase Type not supported")

    @staticmethod
    def __setLocator(lLocatorType, lLocator):
        return utils.setLocatorFromLocatorType(lLocatorType, lLocator)

    @staticmethod
    def __setTimeout(lTimeout):
        return 20 if not lTimeout else float(lTimeout)

    def __doComparisons(self, lComparison, value1, value2):
        if isinstance(value1, bool) or isinstance(value2, bool):
            value1 = utils.anything2Boolean(value1)
            value2 = utils.anything2Boolean(value2)

        if value2 == 'None':
            value2 = None

        if lComparison == "=":
            if value1 == value2:
                self.ifIsTrue = True
            else:
                self.ifIsTrue = False
        elif lComparison == "!=":
            self.ifIsTrue = False if value1 == value2 else True
        elif lComparison == ">":
            if value1 > value2:
                self.ifIsTrue = True
            else:
                self.ifIsTrue = False
        elif lComparison == "<":
            if value1 < value2:
                self.ifIsTrue = True
            else:
                self.ifIsTrue = False
        elif not lComparison:  # Check only, if Value1 has a value.
            self.ifIsTrue = True if value1 else False
        else:
            raise BaseException(f"Comparison Operator not supported/unknown {lComparison}")

        return self.ifIsTrue

    def execute(self):
        """Method is overwritten in all children/subclasses"""
        pass

    def teardown(self):
        if self.testCaseStatus:
            if not self.testcaseDataDict[GC.TESTCASESTATUS]:
                self.testcaseDataDict[GC.TESTCASESTATUS] = self.testCaseStatus
            elif self.testCaseStatus == GC.TESTCASESTATUS_SUCCESS:
                # We'll not overwrite a potential Error Status of the testcase
                pass
            elif self.testCaseStatus == GC.TESTCASESTATUS_ERROR:
                self.testcaseDataDict = GC.TESTCASESTATUS_ERROR
            else:
                sys.exit("No idea, what happened here. Unknown condition appeared")

        self.timing.takeTime(self.timingName)

    def replaceVariables(self, expression):
        """
        The syntax for variables is currently $(<column_name_from_data_file>). Multiple variables can be assigned
        in one cell, for instance perfectly fine: "http://$(BASEURL)/$(ENDPOINT)"

        There's a special syntax for the faker module: $(FAKER.<fakermethod>).

        Also a special syntax for API-Handling: $(APIHandling.<DictElementName>).

        @param expression: the current cell, either as fixed value, e.g. "Franzi" or with a varible $(DATE)
        @return: the replaced value, e.g. if expression was $(DATE) and the value in column "DATE" of data-file was
            "01.01.2020" then return will be "01.01.2020"
        """
        if "$(" not in expression:
            return expression

        while "$(" in expression:
            if expression[0:2] == "$(":
                left_part = ""
            else:
                left_part = expression.split("$(")[0]

            center = expression[len(left_part) + 2:]
            center = center.split(")")[0]

            right_part = expression[len(left_part) + len(center) + 3:]

            if "." not in center:
                # Replace the variable with the value from data structure
                centerValue = self.testcaseDataDict.get(center)
            else:
                # This is a reference to a DICT with ".": for instance APIHandling.AnswerContent("<bla>")
                dictVariable = center.split(".")[0]
                dictValue = center.split(".")[1]

                if dictVariable == 'ANSWER_CONTENT':
                    centerValue = self.apiSession.session[1].answerJSON.get(dictValue, "Empty")
                elif dictVariable == 'FAKER':
                    # This is to call Faker Module with the Method, that is given after the .
                    centerValue = self.__getFakerData(dictValue)
                else:
                    raise BaseException(f"Missing code to replace value for: {center}")

            if not centerValue:
                if center in self.testcaseDataDict.keys():
                    # The variable exists, but has no value.
                    return None
                else:
                    raise BaseException(f"Variable not found: {center}, input parameter was: {expression}")

            expression = "".join([left_part, centerValue, right_part])
        return expression

    def __getFakerData(self, fakerMethod):
        if not self.baangtFaker:
            self.baangtFaker = baangtFaker()

        logger.debug(f"Calling faker with method: {fakerMethod}")

        return self.baangtFaker.fakerProxy(fakerMethod=fakerMethod)




from baangt.base.ProxyRotate import ProxyRotate
from baangt.base import GlobalConstants as GC
from baangt.base.PathManagement import ManagedPaths
import os
import uuid
import logging
from pathlib import Path
import platform
import ctypes
import zipfile
import requests
from dataclasses import dataclass
import tarfile
from urllib.request import urlretrieve

logger = logging.getLogger("pyC")

@dataclass
class BrowserDriverData:
    locatorType : str
    locator : str
    driver : None

class BrowserHelperFunction:
    """
    Helper function class for BrowserHandling. 
    """

    @staticmethod
    def browserHelper_setBrowserDownloadDirRandom():
        """
        Generate a new Directory for downloads. This needs to be specific for each browser session,
        so that we can know, which documents were created during this test case.
        :return:
        """
        randomValue = str(uuid.uuid4())
        downloadFolder = str(Path(ManagedPaths().getOrSetAttachmentDownloadPath()).joinpath(randomValue))
        Path(downloadFolder).mkdir(parents=True, exist_ok=True)

        logger.debug(f"Directory for download {downloadFolder}")
        return downloadFolder

    @staticmethod
    def browserHelper_startBrowsermobProxy(browserName, browserInstance, browserProxy):
        browserProxy.new_har(f"baangt-{browserName}-{browserInstance}",
                             options={'captureHeaders': True, 'captureContent': True}) if browserProxy else None


    @staticmethod
    def browserHelper_setSettingsRemoteV4(desiredCapabilities):
        """
        Extract IP, Port and Browser from the globalSettings, if set.
        If not, use localhost:4444/firefox as default settings

        :param desiredCapabilities: (from globals)
        :return: desiredCapabiligies, IP and Port of Selenium Grid V4
        """
        seleniumGridIp = ""
        seleniumGridPort = ""
        if len(desiredCapabilities) > 0:
            if not 'seleniumGridIp' in desiredCapabilities.keys():
                seleniumGridIp = '127.0.0.1'
            else:
                seleniumGridIp = desiredCapabilities["seleniumGridIp"]
                desiredCapabilities.pop("seleniumGridIp")

            if not 'seleniumGridPort' in desiredCapabilities.keys():
                seleniumGridPort = '4444'
            else:
                seleniumGridPort = desiredCapabilities["seleniumGridPort"]
                desiredCapabilities.pop("seleniumGridPort")

            if not 'browserName' in desiredCapabilities.keys():
                desiredCapabilities['browserName'] = 'firefox'

        return desiredCapabilities, seleniumGridIp, seleniumGridPort

    @staticmethod
    def browserHelper_getBrowserExecutable(browserName):
        # Get executable
        if GC.BROWSER_FIREFOX == browserName:
            executable = GC.GECKO_DRIVER
        elif GC.BROWSER_CHROME == browserName:
            executable = GC.CHROME_DRIVER
        else:
           executable = None 

        if 'NT' not in os.name.upper() and executable:
            executable = executable.split('.')[0]
        return executable

    @staticmethod
    def browserHelper_findBrowserDriverPaths(filename):

        lCurPath = Path(ManagedPaths().getOrSetDriverPath())
        lCurPath = lCurPath.joinpath(filename)

        logger.debug(f"Path for BrowserDrivers: {lCurPath}")
        return str(lCurPath)


    @staticmethod
    def browserHelper_log(logType, logText, browserData, cbTakeScreenshot = None, **kwargs):
        """
        Interal wrapper of Browser-Class for Logging. Takes a screenshot on Error and Warning.

        @param logType: any of logging.ERROR, logging.WARN, INFO, etc.
        @param logText: Text to log
        @param kwargs: Additional Arguments to be logged
        """
        argsString = ""
        xshot = "Couldn't take Screenshot"

        for key, value in kwargs.items():
            if value:
                argsString = argsString + f" {key}: {value}"

        if browserData.locator:
            argsString = argsString + f" Locator: {browserData.locatorType} = {browserData.locator}"

        if logType == logging.DEBUG:
            logger.debug(logText + argsString)
        elif logType == logging.ERROR:
            if cbTakeScreenshot:
                xshot = cbTakeScreenshot()
            logger.error(logText + argsString + f" Screenshot: {xshot}")
        elif logType == logging.WARN:
            logger.warning(logText + argsString)
        elif logType == logging.INFO:
            logger.info(logText + argsString)
        elif logType == logging.CRITICAL:
            if cbTakeScreenshot:
                xshot = cbTakeScreenshot()
            logger.critical(logText + argsString + f" Screenshot: {xshot}")
        else:
            print(f"Unknown call to Logger: {logType}")
            BrowserHelperFunction.browserHelper_log(logging.CRITICAL, f"Unknown type in call to logger: {logType}", browserData, cbTakeScreenshot = cbTakeScreenshot)


    @staticmethod
    def browserHelper_setProxyError(randomProxy):
        """
        Inform the central proxy service, that there was an error. OK, it might have been the page itself, that has
        an error and we'll never know. But more likely it's from the Proxy.
        :return:
        """
        if randomProxy:
            ProxyRotate().remove_proxy(ip=randomProxy["ip"], port=randomProxy["port"], type=randomProxy.get("type"))


    @staticmethod
    def browserHelper_getFirefoxFileUrl():
        #TODO check if this function can be combined with browserHelper_getChromeFileUrl
        # Can ctypes.c_voidp used for both browsers?
        url = None
        response = requests.get(GC.GECKO_URL)
        gecko = response.json()
        gecko = gecko['assets']
        gecko_length_results = len(gecko)
        drivers_url_dict = []

        for i in range(gecko_length_results):
            drivers_url_dict.append(gecko[i]['browser_download_url'])

        isTarFile = True
        zipbObj = zip(GC.OS_list, drivers_url_dict)
        geckoDriversDict = dict(zipbObj)
        if platform.system().lower() == GC.WIN_PLATFORM:
            isTarFile = False
            if ctypes.sizeof(ctypes.c_voidp) == GC.BIT_64:
                url = geckoDriversDict[GC.OS_list[4]]
            else:
                url = geckoDriversDict[GC.OS_list[3]]
        elif platform.system().lower() == GC.LINUX_PLATFORM:
            if ctypes.sizeof(ctypes.c_voidp) == GC.BIT_64:
                url = geckoDriversDict[GC.OS_list[1]]
            else:
                url = geckoDriversDict[GC.OS_list[0]]
        else:
            url = geckoDriversDict[GC.OS_list[2]]

        return url, isTarFile

    @staticmethod
    def browserHelper_getChromeFileUrl():
        url = None
        response = requests.get(GC.CHROME_URL)
        chromeversion = response.text
        chromedriver_url_dict = []

        for i in range(len(GC.OS_list_chrome)):
            OS = GC.OS_list_chrome[i]
            chrome = f'http://chromedriver.storage.googleapis.com/{chromeversion}/chromedriver_{OS}.zip'
            chromedriver_url_dict.append(chrome)

        zipbObjChrome = zip(GC.OS_list, chromedriver_url_dict)
        chromeDriversDict = dict(zipbObjChrome)
        if platform.system().lower() == GC.WIN_PLATFORM:
            url = chromeDriversDict[GC.OS_list[3]]
        elif platform.system().lower() == GC.LINUX_PLATFORM:
            url = chromeDriversDict[GC.OS_list[1]]
        else:
            url = chromeDriversDict[GC.OS_list[2]]

        return url
  
    @staticmethod
    def browserHelper_unzipDriverFile(url, path, driverName):
        file = requests.get(url)
        path_zip = path.joinpath(driverName.replace('exe', 'zip'))
        logger.debug(f"Zipfile with browser expected here: {path_zip} ")
        open(path_zip, 'wb').write(file.content)
        with zipfile.ZipFile(path_zip, 'r') as zip_ref:
            zip_ref.extractall(path)

        if platform.system().lower() != GC.WIN_PLATFORM:
            file_path = path.joinpath(driverName.replace('.exe', ''))
            os.chmod(file_path, 0o777)
        os.remove(path_zip)

    @staticmethod
    def browserHelper_extractTarDriverFile(url, path, driverName):
        path_zip = path.joinpath(driverName.replace('exe', 'tar.gz'))
        filename, _ = urlretrieve(url, path_zip)
        logger.debug(f"Tarfile with browser expected here: {filename} ")
        tar = tarfile.open(filename, "r:gz")
        tar.extractall(path=path)
        tar.close()



import logging
from pathlib import Path
import os
import platform


logger = logging.getLogger("pyC")


class DownloadFolderMonitoring:
    """
    Browserdrivers download all attachments into specific paths
    Here we shall watch for files, that were downloaded in this browser session
    and provide a method to report the last or all files for the current
    test case.

    A browser Driver session can run more than 1 testcase

    """

    def __init__(self, directoryToMonitor):
        self.directoryToWatch = directoryToMonitor
        # Get the current list - before we're started from the directory
        # This should be 0 as we've just created this folder
        self.__files = self.__readFiles()

    def getNewFiles(self):
        """

        :return: List of changed Files in directory since last call
        """
        lFiles = self.__readFiles()

        lNewFiles = [item for item in lFiles if item not in self.__files]

        if lNewFiles:
            logger.debug(f"List of new files in Folder {self.directoryToWatch}: {lNewFiles}")
        else:
            logger.debug(f"no new files detected in Folder {self.directoryToWatch}")

        self.__files = lFiles  # to avoid giving the same new results as on last call.

        return lNewFiles

    def __readFiles(self):
        newFiles = []
        newFilesWithDate = []
        for file in os.scandir(self.directoryToWatch):
            if not file.is_dir():
                newFiles.append(file)

        for file in newFiles:
            lCreationDate = DownloadFolderMonitoring.creation_date(file.path)
            newFilesWithDate.append([file.path, lCreationDate])

        return newFilesWithDate

    @staticmethod
    def creation_date(path_to_file):
        """
        Try to get the date that a file was created, falling back to when it was
        last modified if that isn't possible.
        See http://stackoverflow.com/a/39501288/1709587 for explanation.
        """
        if platform.system() == 'Windows':
            return os.path.getctime(path_to_file)
        else:
            stat = os.stat(path_to_file)
            try:
                return stat.st_birthtime
            except AttributeError:
                # We're probably on Linux. No easy way to get creation dates here,
                # so we'll settle for when its content was last modified.
                return stat.st_mtime




from baangt.base.Utils import utils
from baangt.base.TestRunUtils import TestRunUtils
import baangt.base.GlobalConstants as GC
import baangt.base.CustGlobalConstants as CGC
import xlrd
import logging

logger = logging.getLogger("pyC")


class TestRunExcelImporter:
    """
    The TestrunSettings are in class TestRunUtils and expected to be a deep dict. For details see documentation there.

    This class will migrate data from an excel sheet (either simple format with only 1 tab or complex format with all
    structural elements) into the deep dict.

    In case the XLSX is simple format, all missing data is "predicted"/assumed.

    """
    def __init__(self, FileNameAndPath, testRunUtils: TestRunUtils):
        self.testStepExecutionNumber = 0
        # self.testRunAttributes = testRunUtils.testRunAttributes
        self.testRunUtils = testRunUtils

        try:
            self.excelFile = xlrd.open_workbook(FileNameAndPath)
        except FileNotFoundError as e:
            raise BaseException(f"File not found - exiting {e}")
        self.fileName = utils.extractFileNameFromFullPath(FileNameAndPath)

    def importConfig(self, global_settings):
        self._initTestRun(global_settings)
        return self.testRunUtils.getCompleteTestRunAttributes(self.fileName)

    def _initTestRun(self, global_settings):
        """
        Writes a new entry to testRunUtils --> = this current TestRunDefinition (Filename)
        Then loops through the tabs "TestRun", "ExportFieldList", "TestCaseSequence" and so on,
        reads the lines/columns and writes them the the deep dict of testRunUtils accordingly (TestCaseSequence,
        TestCase, TestStepSequence, TestSteps, etc.)
        @return:
        """
        self.testRunUtils.testRunAttributes = \
            {
                self.fileName: {
                    GC.KWARGS_TESTRUNATTRIBUTES: {
                    },
                },
            }
        testRunAttributes = self.testRunUtils.testRunAttributes[self.fileName][GC.KWARGS_TESTRUNATTRIBUTES]

        xlsTab = self._getTab("TestRun")
        if xlsTab:
            testRunAttributes[GC.EXPORT_FORMAT] = \
                {GC.EXPORT_FORMAT: self._getValueFromList(xlsTab=xlsTab,
                                                          searchField=GC.EXPORT_FORMAT)}
        else:
            # Default Value
            testRunAttributes[GC.EXPORT_FORMAT] = {GC.EXPORT_FORMAT: GC.EXP_XLSX}

        xlsTab = self._getTab("ExportFieldList")
        if xlsTab:
            testRunAttributes[GC.EXPORT_FORMAT][GC.EXP_FIELDLIST] = self._getRowAsList(xlsTab)
        else:
            # Default Value as there are no explicit values from the TestRunDefinition:
            testRunAttributes[GC.EXPORT_FORMAT][GC.EXP_FIELDLIST] = [GC.TESTCASESTATUS,
                                                                     GC.TIMING_DURATION,
                                                                     GC.TIMELOG]

        testRunAttributes[GC.STRUCTURE_TESTCASESEQUENCE] = {}
        testrunSequence = testRunAttributes[GC.STRUCTURE_TESTCASESEQUENCE]

        xlsTab = self._getTab("TestCaseSequence")
        if xlsTab:
            lSequenceDict = self.getRowsWithHeadersAsDict(xlsTab=xlsTab)
        else:
            lSequenceDict = {1:
                {
                "SequenceClass": GC.CLASSES_TESTCASESEQUENCE,
                "TestDataFileName": self.fileName,
                "Sheetname": "data",
                "ParallelRuns": 1,
                "Lines": "0-999999"
                },
            }
        for key, sequence in lSequenceDict.items():
            testrunSequence[key] = [sequence["SequenceClass"], {}]
            for field, value in sequence.items():
                testRunAttributes[GC.STRUCTURE_TESTCASESEQUENCE][key][1][field] = value


        xlsTab = self._getTab("TestCase")
        # if Tab "TestCase" exists, then take the definitions from there. Otherwise (means simpleFormat)
        # we need to create "dummy" data ourselves.
        if xlsTab:
            lTestCaseDict = self.getRowsWithHeadersAsDict(xlsTab=xlsTab)
        else:
            # Dirty, but not possible in any other way in API-Simple-Format:
            if "API" in self.fileName.upper():
                    lTestCaseDict = {1: {"TestCaseSequenceNumber": 1,
                                     "TestCaseNumber": 1,
                                     "TestCaseType": GC.KWARGS_API_SESSION,
                                     "TestCaseClass": GC.CLASSES_TESTCASE}}
            # If there's no defaults in the testrun definition, we need to make some assumptions:
            else:
                if global_settings.get('TC.Mobile'):
                    lTestCaseDict = {1: {"TestCaseSequenceNumber": 1,
                                         "TestCaseNumber": 1,
                                         "TestCaseType": GC.KWARGS_BROWSER,
                                         "TestCaseClass": GC.CLASSES_TESTCASE,
                                         GC.KWARGS_BROWSER: GC.BROWSER_FIREFOX,
                                         GC.BROWSER_ATTRIBUTES: "",
                                         GC.KWARGS_MOBILE: "",
                                         GC.KWARGS_MOBILE_APP: "",
                                         GC.MOBILE_PLATFORM_NAME: "",
                                         GC.MOBILE_DEVICE_NAME: "",
                                         GC.MOBILE_PLATFORM_VERSION: "",
                                         GC.MOBILE_APP_URL: "",
                                         GC.MOBILE_APP_PACKAGE : "",
                                         GC.MOBILE_APP_ACTIVITY : ""}}

                else:
                    lTestCaseDict = {1: {"TestCaseSequenceNumber": 1,
                                     "TestCaseNumber": 1,
                                     "TestCaseType": GC.KWARGS_BROWSER,
                                     "TestCaseClass": GC.CLASSES_TESTCASE,
                                     GC.KWARGS_BROWSER: GC.BROWSER_FIREFOX,
                                     GC.BROWSER_ATTRIBUTES: "",
                                     GC.BROWSER_WINDOW_SIZE: "1900x1200"}}
        # Now we either have data by assumptions or from the testun definition. We'll take it into internal format:
        for key, testCase in lTestCaseDict.items():
            testSequenceRoot = testrunSequence[testCase["TestCaseSequenceNumber"]][1]
            testSequenceRoot[GC.STRUCTURE_TESTCASE] = {testCase["TestCaseNumber"]: []}
            testSequenceRoot[GC.STRUCTURE_TESTCASE][testCase["TestCaseNumber"]].append(testCase["TestCaseClass"])
            testSequenceRoot[GC.STRUCTURE_TESTCASE][testCase["TestCaseNumber"]].append(
                {"TestCaseType": testCase["TestCaseType"],
                 GC.KWARGS_BROWSER: testCase.get(GC.KWARGS_BROWSER),
                 GC.BROWSER_ATTRIBUTES: testCase.get(GC.BROWSER_ATTRIBUTES),
                 GC.BROWSER_WINDOW_SIZE: testCase.get(GC.BROWSER_WINDOW_SIZE),
                 GC.KWARGS_MOBILE: testCase.get(GC.KWARGS_MOBILE),
                 GC.KWARGS_MOBILE_APP: testCase.get(GC.KWARGS_MOBILE_APP),
                 GC.MOBILE_PLATFORM_NAME: testCase.get(GC.MOBILE_PLATFORM_NAME),
                 GC.MOBILE_DEVICE_NAME: testCase.get(GC.MOBILE_DEVICE_NAME),
                 GC.MOBILE_PLATFORM_VERSION: testCase.get(GC.MOBILE_PLATFORM_VERSION),
                 GC.MOBILE_APP_URL: testCase.get(GC.MOBILE_APP_URL),
                 GC.MOBILE_APP_PACKAGE: testCase.get(GC.MOBILE_APP_PACKAGE),
                 GC.MOBILE_APP_ACTIVITY: testCase.get(GC.MOBILE_APP_ACTIVITY)
                 })

        xlsTab = self._getTab("TestStep")
        if xlsTab:
            lStepDict = self.getRowsWithHeadersAsDict(xlsTab=xlsTab)
        else:
            lStepDict = {1: {"TestCaseSequenceNumber": 1,
                         "TestCaseNumber": 1,
                         "TestStepNumber": 1,
                         "TestStepClass": GC.CLASSES_TESTSTEPMASTER}
                         }
        for key, testStep in lStepDict.items():
            testStepRoot = testrunSequence[testStep["TestCaseSequenceNumber"]][1]
            testStepRoot = testStepRoot[GC.STRUCTURE_TESTCASE][testStep["TestCaseNumber"]]
            if len(testStepRoot) == 2:
                testStepRoot.append({GC.STRUCTURE_TESTSTEP: {}})

            testStepRoot = testStepRoot[2][GC.STRUCTURE_TESTSTEP]
            testStepRoot[testStep["TestStepNumber"]] = [{"TestStepClass": testStep["TestStepClass"]},
                                                        {GC.STRUCTURE_TESTSTEPEXECUTION: {}}]


        xlsTab = self._getTab("TestStepExecution")
        lExecDict = self.getRowsWithHeadersAsDict(xlsTab=xlsTab)
        for key, execLine in lExecDict.items():
            lSequence = self.testRunUtils.getSequenceByNumber(testRunName=self.fileName,
                                                              sequence=execLine.get("TestCaseSequenceNumber",1))
            lTestCase = self.testRunUtils.getTestCaseByNumber(sequence=lSequence,
                                                              testcaseNumber=execLine.get("TestCaseNumber",1))
            lTestStep = self.testRunUtils.getTestStepByNumber(testCase=lTestCase,
                                                              testStepNumber=execLine.get("TestStepNumber",1))
            # if this TestStep is still just a String, then it needs to be converted into a List with Dict
            # to hold the TestexecutionSteps in this Dict.
            if isinstance(lTestStep, str):
                lTestCase[2][GC.STRUCTURE_TESTSTEP][execLine.get("TestStepNumber",1)] = \
                    [lTestStep, {GC.STRUCTURE_TESTSTEPEXECUTION: {}}]
                lTestStep = self.testRunUtils.getTestStepByNumber(testCase=lTestCase,
                                                                  testStepNumber=execLine.get("TestStepNumber",1))

            lStepExecutionNumber = self.__iterateStepExecutionNumber(execLine.get("TestStepExecutionNumber"))
            lTestStep[1][GC.STRUCTURE_TESTSTEPEXECUTION][lStepExecutionNumber] = {}
            for lkey, value in execLine.items():
                lTestStep[1][GC.STRUCTURE_TESTSTEPEXECUTION][lStepExecutionNumber][lkey] = value

    def __iterateStepExecutionNumber(self, numberFromDefinition):
        if numberFromDefinition:
            return numberFromDefinition

        # This is a teststep in simple format, without an execution number column in the XLS
        # we need to iterate ourselves and simply add 1 to each new TestStep
        self.testStepExecutionNumber += 1
        return self.testStepExecutionNumber

    def getRowsWithHeadersAsDict(self, xlsTab):
        lRetDict = {}

        for row in range(1, xlsTab.nrows):
            lRetDict[row] = {}
            for col in range(0, xlsTab.ncols):
                lRetDict[row][xlsTab.cell_value(0, col)] = self.replaceFieldValueWithValueOfConstant(
                    xlsTab.cell_value(row, col))
        return lRetDict

    def _getRowAsList(self, xlsTab, colNumber=0):
        l_fields = []
        for row in range(1, xlsTab.nrows):
            l_fields.append(self.replaceFieldValueWithValueOfConstant(xlsTab.cell_value(row, colNumber)))
        return l_fields

    def _getValueFromList(self, xlsTab, searchField):
        # Searches for Value in Column 1 and returns Value from Column 2 if found.
        # This is used for XLSX-Tabs with this format:
        # column:value
        # TestlineStart:10
        # TestlineEnd:20
        # foo:bar
        for row in range(1, xlsTab.nrows):
            if xlsTab.cell_value(row, 1) == searchField:
                return self.replaceFieldValueWithValueOfConstant(xlsTab.cell_value(row, 2))

    def _getTab(self, tabName):
        try:
            worksheet = self.excelFile.sheet_by_name(tabName)
        except Exception as e:
            return None
        return worksheet

    def replaceFieldValueWithValueOfConstant(self, value):
        """
        baangt Global constants (baangt.base.GlobalConstants) are available everywhere as GC., e.g. the variable
        "BROWSER" defined in GlobalConstants can be accessed from everywhere within baangt by using "GC.BROWSER".

        The variables can also be used in configuration files (would be very stupid if we need to change a constant and
        then 100s of Config-Files need adjustment - worst case the testruns behave unexepected). This applies both to
        XLSX and JSON Config files.

        The CGC.-Part is still needs fixing. It shouldn't appear in Baangt base, but currently still needed.

        @param value: potentially convertable value (e.g. GC.BROWSER)
        @return: potentially converted value (e.g. "Browser")
        """
        if isinstance(value, float):
            if value % 1 == 0:
                return int(value)
        elif isinstance(value, int):
            return value

        value = value.strip()
        value = utils.replaceFieldValueWithValueOfConstant(value)

        if value[0:5] == "CGC.":
            value = getattr(globals()[value.split(".")[0]], value.split(".")[1])

        return value



import json
import os
# This Python file uses the following encoding: utf-8


class GlobalSettings:
    """ A Singleton Class to hold all settings in Baant.
    Read globalSettings.json file and prepare dictonary data.
    Method:
        getInstance : Public static method to get return Instance object
        saveSimpleJson:  Save settings as json file
        returnSetting:  Return dictionary of all settings

    JsonFile Each Entry Format:
        "variableName":{ Actual Variable Name used in program
                        "hint ": This information will be shown as tooltip
                        "type":  based on input, Form field created
                                 bool : Checkbox
                                 type : TextInputBox
                                 select :  ComboBox
                        "default":  Default value if nothing is provided
                        "options" :  List Type for all type of possible
                         "display" : This will be displayed
                      }

    New Changes: As per Discussion:
    New variable self.globalconfig is used to store setting of
    globalSetting.json and self.config will be used to store
    combined setting from global as well as config FileExistsError
    """
    __instance = None

    class __GlobalSettings:
        """ Inner Class which will be used """

        def __init__(self, jsonfile=None):
            """ Initialize the config data from json file """
            self.config = {}
            self.globalconfig = {}
            if jsonfile:
                # internally call initBaseConfig and addValue
                self.addValue(jsonfile)
            else:
                self.initBaseConfig()

        def initBaseConfig(self):
            """ This function will initialize the config data
            with globalSetting.json content
            """
            # init the file to base state
            if self.config:
                self.config = {}

            globalFile = "globalSetting.json"
            if not os.path.isfile(globalFile):
                dirpath = os.path.dirname(os.path.abspath(__file__))
                globalFile = os.path.join(dirpath, 'globalSetting.json')
            if not os.path.isfile(globalFile):
                # package attempt is failed
                dirpath = os.getcwd()
                globalFile = os.path.join(os.path.dirname(dirpath), 'globalSetting.json')
            # alert on console if globalSettings not found
            if not os.path.isfile(globalFile):
                print(" {} File not a valid file ".format(globalFile))
            data = self.parseJsonfile(globalFile)
            # self.config = data
            self.globalconfig = data


        def isValidFile(self, jsonfile):
            """ This return True if it can be opened """
            try:
                with open(jsonfile) as f:
                    json.loads(f.read())
                return True
            except Exception as e:
                print("Unable to read the file {}".format(jsonfile))
                print(e)
                return False

        def filterIniKey(self, filterkey=[]):
            """ This function will remove
            RootPath, ScreenshotPath, ImportPath, ExportPath
            """
            data = {}
            iniData = {'RootPath', 'ScreenshotPath',
                       'ImportPath', 'ExportPath'}
            if filterkey:
                iniData.update(filterkey)

            for key, value in self.config.items():
                if key not in iniData:
                    data[key] = value

            return data

        def parseJsonfile(self, jsonfile):
            """ This function will parse the input json file,
            and return  formated as per baangt
            Accept Format:
                "Release":{'hint':"somehint','type':'bool',
                            'options':['some optins'],
                             'displayText':"shown"
                             }
                 "Release":"22.21.b.dev"
            """
            if self.isValidFile(jsonfile):
                # loop with each variable
                # empty dictionary to store the parse valued
                resultDict = {}
                with open(jsonfile) as f:
                    data = json.loads(f.read())
                    # check if its globalSetting.json
                    if data.get('settings'):
                        # parse the global file
                        resultDict.update(data['settings'])
                    else:
                        for key, value in data.items():
                            resultDict[key] = GlobalSettings.transformToDict(key, value)

                return resultDict
            else:
                print("Invalid Format ")
                return {}

        def addValue(self, jsonfile):
            """ This will update the existing settings with given dictionary
            of data in json file.
            Also, this will remove all current dictionary data

            """
            # init with base config and store variable in globalconfig
            self.initBaseConfig()

            # path should be absolute to process the file
            parsed_data = self.parseJsonfile(jsonfile)
            if parsed_data:
                for key, value in parsed_data.items():
                    if key in self.globalconfig:
                        # update globalconfig and return value to config
                        # self.config[key] = self.globalconfig[key]
                        # update the value
                        #self.config[key]['default'] = value['default']
                        # changes: store the default value
                        self.config[key] = value['default']
                    else:
                        self.config[key] = value

        def updateValue(self, dictData):
            """ This function will update existing Dictionary setting
            from the provided value
            """
            if isinstance(dictData, dict):
                # loop each items
                # No need to loop just update
                self.config.update(dictData)
                # for key in dictData:
                #   if key in self.config:
                #       # The key is present so update the value
                #       self.config[key]['default'] = dictData[key]
                #   else:
                #       # it will be user defined key, will be text
                #       self.config[key] = GlobalSettings.transformToDict(
                #                               key,
                #                               dictData[key])

            # if key in self.confg not found in dictData
            # delete button is pressed, need to remove from self.config also
            for key in list(self.config):
                if key not in dictData:
                    self.config.pop(key)

    def __new__(cls, jsonfile=None):
        """ Get the instance and update the value """
        if not GlobalSettings.__instance:
            # Create New instance
            GlobalSettings.__instance = GlobalSettings.__GlobalSettings(
                                     jsonfile
                                     )
        else:
            GlobalSettings.__instance.addValue(jsonfile)
        return GlobalSettings.__instance

    @staticmethod
    def getInstance():
        """ Return current instance of Class"""
        if GlobalSettings.__instance is None:
            # Create a new instance
            GlobalSettings.__instance = GlobalSettings.__GlobalSettings()
        return GlobalSettings.__instance

    @staticmethod
    def transformToDict(key, value):
        """ this convert value to dictionary format with keys
        "hint" : default to value
        "type" : default to "text"
        "default":  default to value
        "options" :  default to value
        "diplayText" :  default to key
        Example:  Input = "path"
                  Output = {'hint':'path',
                            'type': 'text',
                            'default': 'path',
                            'options': ['path'],
                            'displyText': 'key'
                            }
        """
        data = {}
        if isinstance(value, dict):
            # dictionary item, get the require parameter
            defaultValue = value.get('default','sampleValue')
            data['hint'] = value.get('hint', defaultValue)
            data['type'] = value.get('type', 'text')
            data['options'] = value.get('options', [defaultValue])
            data['displayText'] = value.get('displayText', key)
            data['default'] = defaultValue

            # return data
            return data
        elif isinstance(value, str):
            value = value
        elif isinstance(value, set):
            value = value.pop()
        elif isinstance(value, bool):
            value = str(value)
        else:
            print("Invalid value  : {}".format(value))
            value = "exampleValue"
        data['hint'] = value
        data['type'] = 'text'
        data['options'] = [value]
        data['displayText'] = key
        data['default'] = value

        return data


if __name__ == "__main__":
    d = GlobalSettings('sample.json')
    print("After adding new file")
    print("Data in config ", d.config)
    print("Data in globalconfig ", d.globalconfig)

    print("Initializing the data")
    d = GlobalSettings('globalSetting.json')
    print("if no instance earlier")
    print("Data in config ", d.config)
    print("Data in globalconfig ", d.globalconfig)



KWARGS_DATA = "data"
KWARGS_MOBILE = "Mobile"
KWARGS_MOBILE_APP = "MobileApp"
KWARGS_BROWSER = "Browser"
KWARGS_API_SESSION = "api"
KWARGS_APPIUM = 'Appium'
KWARGS_TESTCASETYPE = "TestCaseType"
KWARGS_TESTRUNATTRIBUTES = "TESTRUNEXECUTIONPARAMETERS"
KWARGS_TESTRUNINSTANCE = "TESTRUNINSTANCE"
KWARGS_TIMING = "TimingClassInstance"
KWARGS_SEQUENCENUMBER = "SequenceNumberOfThisTestCase"

CLASSES_TESTCASESEQUENCE = "baangt.TestCaseSequence.TestCaseSequenceMaster.TestCaseSequenceMaster"
# CLASSES_TESTCASESEQUENCE_new = "TestCaseSequenceMaster"
CLASSES_TESTCASE = "baangt.TestCase.TestCaseMaster.TestCaseMaster"
# CLASSES_TESTCASE_new = 'TestCaseMaster'
CLASSES_TESTSTEPMASTER = 'baangt.TestSteps.TestStepMaster.TestStepMaster'
# CLASSES_TESTSTEPMASTER_new = 'TestStepMaster'

TIMING_END = "end"
TIMING_START = "start"
TIMING_TESTRUN = "Complete Testrun"
TIMING_DURATION = "Duration"
TIMESTAMP = "timestamp"
TIMELOG = "timelog"

GECKO_DRIVER = "geckodriver.exe"
CHROME_DRIVER = "chromedriver.exe"
EDGE_DRIVER = "msedgedriver.exe"
REMOTE_EXECUTE_URL = "http://localhost:4444/wd/hub"

BROWSER_FIREFOX = "FF"
BROWSER_CHROME = "CHROME"
BROWSER_SAFARI = "SAFARI"
BROWSER_EDGE = "EDGE"
BROWSER_REMOTE = 'REMOTE'
BROWSER_REMOTE_V4 = 'REMOTE_V4'
BROWSER_APPIUM = 'APPIUM'
BROWSER_MODE_HEADLESS = "HEADLESS"
BROWSER_ATTRIBUTES = "BrowserAttributes"
BROWSER_WINDOW_SIZE = "BrowserWindowSize"
BROWSER_ZOOM_FACTOR = "BrowserZoomFactor"

CMD_CLICK = "CLICK"
CMD_SETTEXT = "SETTEXT"
CMD_FORCETEXT = "FORCETEXT"

TESTCASESTATUS = "TestCaseStatus"
TESTCASESTATUS_SUCCESS = "OK"
TESTCASESTATUS_ERROR = "Failed"
TESTCASESTATUS_WAITING = "Paused"
TESTCASEERRORLOG = "TCErrorLog"
TESTCASESTATUS_STOP = "TCStopTestCase"
TESTCASESTATUS_STOPERROR = "TCStopTestCaseError"
TESTCASE_EXPECTED_ERROR_FIELD = "TC Expected Error"

DATABASE_FROM_LINE = "FromLine"
DATABASE_TO_LINE = "ToLine"
DATABASE_LINES = "Lines"
DATABASE_FILENAME = "TestDataFileName"
DATABASE_SHEETNAME = "Sheetname"
DATABASE_EXPORTFILENAMEANDPATH = "exportFilesBasePath"

STRUCTURE_TESTCASESEQUENCE = "TESTSEQUENCE"
STRUCTURE_TESTCASE = "TESTCASE"
STRUCTURE_TESTSTEP = "TestStep"
STRUCTURE_TESTSTEPEXECUTION = "TestStepExecutionParameters"

EXECUTION_PARALLEL = "ParallelRuns"
SCREENSHOTS = "Screenshots"
EXECUTION_DONTCLOSEBROWSER = "dontCloseBrowser"
EXECUTION_SLOW = "slowExecution"
EXECUTION_NETWORK_INFO = 'NetworkInfo'
EXECUTION_LOGLEVEL = "LogLevel"

EXECUTION_STAGE = 'Stage'
EXECUTION_STAGE_TEST = 'Test'
EXECUTION_STAGE_DEVELOPMENT = 'Development'
EXECUTION_STAGE_PRODUCTION = 'Production'
EXECUTION_STAGE_QA = 'Quality Assurance'

EXPORT_FORMAT = "ExportFormat"
EXPORT_ADDITIONAL_DATA = "AdditionalExportTabs"
EXP_FIELDLIST = "Fieldlist"
EXP_XLSX = "XLSX"
EXP_CSV = "CSV"

PATH_EXPORT = '1TestResults'
PATH_IMPORT = '0TestInput'
PATH_SCREENSHOTS = 'Screenshots'
PATH_ROOT = 'RootPath'

ADDRESS_COUNTRYCODE = "CountryCode"
ADDRESS_POSTLCODE = "PostlCode"
ADDRESS_CITYNAME = "CityName"
ADDRESS_STREETNAME = "StreetName"
ADDRESS_HOUSENUMBER = "HouseNumber"
ADDRESS_ADDITION1 = "Addition1"
ADDRESS_ADDITION2 = "Addition2"

MOBILE_PLATFORM_NAME = "platformName"
MOBILE_DEVICE_NAME = "deviceName"
MOBILE_PLATFORM_VERSION = "platformVersion"
MOBILE_APP_URL = 'app'
MOBILE_APP_PACKAGE = 'appPackage'
MOBILE_APP_ACTIVITY = 'appActivity'
MOBILE_APP_BROWSER_PATH = 'mobileAppBrowserPath'   # Path to Browser on Mobile device

WIN_PLATFORM = 'windows'
LINUX_PLATFORM = 'linux'

BIT_64 = 8
BIT_32 = 4

OS_list = ["Linux-32", "Linux-64", "MacOS", "Windows-32", "Windows-64"]
OS_list_chrome = ['linux32', 'linux64', 'mac64', 'win32']

GECKO_URL = 'https://api.github.com/repos/mozilla/geckodriver/releases/latest'
CHROME_URL = 'https://chromedriver.storage.googleapis.com/LATEST_RELEASE'


BROWSER_PROXY_PATH = '/browsermob-proxy/bin/browsermob-proxy'
BROWSER_PROXY_URL = 'https://github.com/lightbody/browsermob-proxy/releases/download/browsermob-proxy/browsermob-proxy-bin.zip'

PROXY_FAILCOUNTER = 3

TESTDATAGENERATOR_OUTPUT_FORMAT = "xlsx"
TESTDATAGENERATOR_INPUTFILE = "RawTestData.xlsx"
TESTDATAGENERATOR_OUTPUTFILE_XLSX = "output.xlsx"
TESTDATAGENERATOR_OUTPUTFILE_CSV = "output.csv"



import os
import uuid
import xlsxwriter
import glob
import logging
import sys
from bs4 import BeautifulSoup as bs4
from pathlib import Path

def setupLogger():
    import logging
    import os
    from datetime import datetime

    logPath = "/".join(os.path.dirname(os.path.realpath(__file__)).split('/')[0:-1]) + "/logs/"
    Path(logPath).mkdir(parents=True, exist_ok=True)
    logFilename = (logPath +
                   datetime.now().strftime("%Y%m%d_%H%M%S") + '.log')
    print(f"Logfile verwendet: {logFilename}")

    # Bit more advanced logging:
    logger = logging.getLogger('pyC')
    logger.setLevel(logging.DEBUG)
    # create file handler which logs even debug messages
    fileHandler = logging.FileHandler(logFilename, encoding="UTF-8")
    fileHandler.setLevel(level=logging.DEBUG)
    # create console handler with a higher log level
    channelHandler = logging.StreamHandler()
    channelHandler.setLevel(logging.INFO)
    # create formatter and add it to the handlers
    formatter = logging.Formatter('%(asctime)s _ %(levelname)s _ %(module)s _ %(funcName)s : %(message)s')
    channelHandler.setFormatter(formatter)
    fileHandler.setFormatter(formatter)
    # add the handlers to logger
    logger.addHandler(channelHandler)
    logger.addHandler(fileHandler)
    return logger

def readXMLFile(fileNameAndPath):
    with open(fileNameAndPath, "r") as file:
        content=file.readlines()
        content = "".join(content)
        fileAsDOM = bs4(content, "lxml")
    return fileAsDOM

def readFile(fileNameAndPath):
    with open(fileNameAndPath, "r") as file:
        content=file.readlines()
    return content

class LocatorObjects:
    def __init__(self):
        self.objects = {}

    def add(self, objectDefinition, path):
        self.objects[path] = objectDefinition

    def exportXLS(self, wsheet: xlsxwriter.Workbook.worksheet_class):
        wsheet.write(0,0, "Location")
        wsheet.write(0,1, "SelectorType")
        wsheet.write(0,2, "SelectorValue")
        wsheet.write(0,3, "IFrame")
        for line, (lObjectKey, lObjectValue) in enumerate(self.objects.items()):
            wsheet.write(line+1, 0, lObjectValue.fileNameAndPathKatalonInternal)
            wsheet.write(line+1, 1, lObjectValue.selector)
            wsheet.write(line+1, 2, lObjectValue.selectorValue)
            wsheet.write(line+1, 3, lObjectValue.iframe)

class Groovy:
    def __init__(self):
        self.groovyScriptClasses = {}

    def add(self, groovyScript, path):
        self.groovyScriptClasses[path] = groovyScript

    def doReplacementOfLiterals(self, objects: LocatorObjects):
        for key, value in self.groovyScriptClasses.items():
            value.replaceLiteralsWithLocators(objects)


class TestCases:
    def __init__(self):
        self.testcases = {}

    def add(self, testCase, path):
        self.testcases[path] = testCase

    @staticmethod
    def decodeHex(string):
        myUuid = uuid.UUID(string)
        return myUuid.fields

class fileHandling:
    def __init__(self, fileNameAndPath, format='XML'):
        self.fileNameAndPath = fileNameAndPath
        if format == 'XML':
            self.fileAsDOM: bs4 = readXMLFile(fileNameAndPath)
        else:
            self.fileAsDOM = readFile(fileNameAndPath)

        self.name = None
        self.fileNameAndPathKatalonInternal = fileHandling._makeShorterPath(self.fileNameAndPath)

    def logFileContentsHeader(self):
        logger.info(f"Processing file: {self.fileNameAndPath}")

    @staticmethod
    def _makeShorterPath(longPath):
        shortPath = fileHandling.__cutBefore(longPath, "Object Repository/")
        shortPath = fileHandling.__cutBefore(shortPath, "Test Cases/")
        shortPath = fileHandling.__cutBefore(shortPath, "Scripts/")
        return shortPath

    @staticmethod
    def __cutBefore(path, cutBeforeValue):
        if cutBeforeValue in path:
            return cutBeforeValue + "".join(path.split(cutBeforeValue)[1:])
        return path

class translateTestCase(fileHandling):
    def __init__(self, fileNameAndPath):
        super().__init__(fileNameAndPath)
        self.scripts = {}
        self.outputAnalysis()

    def outputAnalysis(self):
        lAttrib = self.fileAsDOM.find("testcaseguid").get_text(strip=True)
        hexDecode = TestCases.decodeHex(lAttrib)
        super().logFileContentsHeader()
        logger.info(f"TestCaseGuid: {lAttrib}, UUID-Decoded: {hexDecode}")
        self.findGroovyScript()

    def findGroovyScript(self):
        # There should be a groovy script with a random number in a directory
        # /Scripts/<path_to_Testcase>/<testCase>/<random>.groovy
        # Current fileNameAndPath: '/Users/bernhardbuhl/git/KatalonVIG/Test Cases/000 Login/Open_VN_Auskunft.tc'
        splitList = self.fileNameAndPath.split("/Test Cases/")
        basePath = splitList[0]
        restPath = "".join(splitList[1:])
        testCaseFileName = "".join(restPath.split("/")[-1:])
        restPath = restPath.replace(testCaseFileName, "")
        testCaseFileName = testCaseFileName.replace(".tc", "")
        searchScriptPath = basePath + "/Scripts/" + restPath + testCaseFileName + "/"
        os.chdir(searchScriptPath)
        l_files = glob.glob("*.groovy")
        for file in l_files:
            self.scripts[file] = translateGoovy(searchScriptPath + file)
            self.scripts[file].savePythonFile(self.fileNameAndPath)

class translateGoovy(fileHandling):
    def __init__(self, fileNameAndPath):
        super().__init__(fileNameAndPath, format="groovy")
        self.interpretGroovy()
        # self.savePythonFile() will be executed from caller

    def interpretGroovy(self):
        lOutput = ""
        for groovyLine in self.fileAsDOM:
            groovyLine = translateGoovy.replaceGroovyLine(groovyLine)
            if len(groovyLine) > 0:
                lOutput += groovyLine
        self.fileAsDOM = lOutput
        self.replaceLiteralsWithLocators("Dummy")

    def replaceLiteralsWithLocators(self, locatorObjectClasses: LocatorObjects):
        """Replace the internal Katalon-Links with actual definitions"""
        # The internal Locator is '<ObjectFolder>/<subfolder./.n>/<name>
        # The filename is '/Object Repository/<ObjectFolder>/<subfolder./.n>/<name>.rs
        basePath = self.fileNameAndPath.split("/Scripts/")[0]
        lOutput = ""
        lInput = self.fileAsDOM.split("\n")
        for line in lInput:
            # fixme: Here's an error - removes variable names
            if "fragen" in self.fileAsDOM:
                x = 2
            if "(xpath='" in line:
                lFileAndPathLocator = line.split("xpath='")[1]
                lFileAndPathLocator = lFileAndPathLocator.split("'")[0]
                lFileAndPathLocator = basePath + "/Object Repository/" + lFileAndPathLocator + ".rs"
                # Sometimes Katalon Studio prefixes with Object Repository and sometimes it doesn't.
                lFileAndPathLocator = lFileAndPathLocator.replace("/Object Repository/Object Repository/", "/Object Repository/")
                lLocatorObject = translateObjectDefinition(lFileAndPathLocator)

                lStart = line.split("xpath='")[0]
                lEnd = line.split("xpath='")[1:]
                lEnd = "'".join(lEnd[0].split("'")[1:])

                if lLocatorObject.selectorValue:
                    lLocatorObject.selectorValue = lLocatorObject.selectorValue.replace("'", '"')
                    line = lStart + lLocatorObject.selector.lower() + "='" + lLocatorObject.selectorValue + "'" + lEnd
                else:
                    logger.warning(f"Couldn't find locator definition for Object {lLocatorObject.fileNameAndPathKatalonInternal}")
                    line = "# FIXME: " + lStart + lLocatorObject.fileNameAndPathKatalonInternal + lEnd

            lOutput = lOutput + "\n" + line
        self.fileAsDOM = lOutput

    def savePythonFile(self, saveAsClassName = None):
        if saveAsClassName:
            lFileExport = saveAsClassName
            lFileExport = lFileExport.replace(".tc", ".py")
            self._createHeaderForClass(saveAsClassName)
        else:
            lFileExport = self.fileNameAndPath.replace(".groovy", ".py")
        lFileExport = self._replacePathForExport(lFileExport)
        lPathExport = "/".join(lFileExport.split("/")[0:-1])
        Path(lPathExport).mkdir(parents=True, exist_ok=True)
        with open(lFileExport, "w") as file:
            file.writelines(self.fileAsDOM)

    @staticmethod
    def _replacePathForExport(fullPathAndFileName):
        lFileExport = fullPathAndFileName.replace("/Test Cases/", "/Python/Test Cases/")
        lFileExport = lFileExport.replace("/Scripts/", "/Python/Scripts/")
        if not "Python" in lFileExport:
            logger.critical(f"Should replace path, but can't. input: {fullPathAndFileName}. \nI better stop now.")
            sys.exit("Critical program error or so.")
        return lFileExport

    def _createHeaderForClass(self, classNameForHeader):
        header = """
import baangt.base.GlobalConstants as GC
from baangt.TestSteps.TestStepMaster import TestStepMaster


"""
        header = header + f"class {classNameForHeader.split('/')[-1].replace('.tc', '')}(TestStepMaster):"

        header = header + """
\tdef __init__(self, **kwargs):
\t\tsuper().__init__(**kwargs)
\t\tself.execute()

\tdef execute(self):"""

        self.addTabs()
        self.fileAsDOM = header + '\n' + self.fileAsDOM

    def addTabs(self):
        """
        Add Tab-Stop to each from Groovy converted Code-Line as this code runs under Method execute()
        """
        self.fileAsDOM = self.fileAsDOM.replace("\n\n", "\n")
        l_lines = self.fileAsDOM.split("\n")
        result = ""
        for line in l_lines:
            line = '\t\t' + line
            result = result + '\n' + line
        self.fileAsDOM = result

    @staticmethod
    def replaceGroovyLine(l_string):
        l_string = l_string.replace(") {", "")
        l_string = l_string.replace("WebUI.click(findTestObject(", "self.browserSession.findByAndClick(xpath=")
        l_string = l_string.replace("WebUI.verifyElementPresent(findTestObject(", "self.browserSession.findBy(xpath=")
        l_string = l_string.replace("WebUI.setText(findTestObject(", "self.browserSession.findByAndSetText(xpath=")
        l_string = l_string.replace("WebUI.waitForElementVisible(findTestObject(", "self.browserSession.findBy(xpath=")
        l_string = l_string.replace("WebUI.verifyElementNotPresent(findTestObject(",
                                    "self.browserSession.findWaitNotVisible(xpath=")
        l_string = l_string.replace("WebUI.waitForElementNotPresent(findTestObject(",
                                    "self.browserSession.findWaitNotVisible(xpath=")
        l_string = l_string.replace("WebUI.doubleClick(findTestObject(",
                                    "self.browserSession.findByAndClick(xpath=")
        l_string = l_string.replace("WebUI.waitForElementPresent(findTestObject(",
                                    "self.browserSession.findBy(xpath=")
        l_string = l_string.replace("if (", "if ")
        l_string = l_string.replace(" && ", " and ")
        l_string = l_string.replacE(" || ", " or ")
        l_string = l_string.replace("else", "else:")
        l_string = l_string.replace("{", "")
        l_string = l_string.replace("}", "")
        l_string = l_string.replace("))", ")")
        l_string = l_string.replace(".toString()","")
        l_string = l_string.replace(".trim()", ".strip()")

        if "WebUI.waitForElementClickable" in l_string:
            l_string = ""

        if "//" in l_string.strip()[0:2]:
            l_string = l_string.replace("//", "# ")

        if "'" in l_string.strip()[0:2]:
            l_string = "# " + l_string

        if "not_run" in l_string.strip()[0:7]:
            l_string = l_string.replace("not_run:", "# ")

        if "if " in l_string:
            l_string = l_string.strip() + ":\n"

        if "import " in l_string[0:10]:
            l_string = ""

        if "Thread.sleep" in l_string:
            sleep_time = l_string.split("(")[1]
            sleep_time = sleep_time.split(")")[0]
            # Sleep time in Milliseonds.
            sleep_time = float(sleep_time) / 1000
            l_string = l_string.split("Thread")[0] + "self.browserSession.sleep(" + str(sleep_time) + ")\n"

        if "WebUI.delay(" in l_string:
            sleep_time = l_string.split("(")[1]
            sleep_time = sleep_time.split(")")[0]
            # sleep time in Seconds
            l_string = l_string.split("WebUI")[0] + "self.browserSession.sleep(" + sleep_time + ")\n"

        if "scrollTo" in l_string:
            l_string = ""

        if "CustomKeyword" in l_string:
            l_string = ""

        if "self.testcaseDataDict" in l_string and "self.browserSession" in l_string and l_string[-1:] == ")":
            l_string = l_string[0:-1] + "'])"
            # schließende Klammer zu früh...
            l_line_split = l_string.split(")")
            l_string = l_line_split[0] + l_line_split[1] + ")"
            # value als named parameter übergeben
            l_string = l_string.replace("self.testcaseDataDict", "value=self.testcaseDataDict")

        l_string = translateGoovy._replaceGlobalVariable(l_string)
        l_string = translateGoovy._replaceOtherStuff(l_string)
        return l_string

    @staticmethod
    def _replaceOtherStuff(stringIn):
        if ".contains(" in stringIn:
            # Syntax: "<space><bla>.contains('blabla')<space>
            # fixme: not done yet. Implement...
            pass
        if "self.browserSession.findByAndSetText(" in stringIn:
            # in self.browserSession.findByAndSetText( after the xpath='value' there is a closing bracket before the value
            # That must got and self.testcasedataDict must be preceeded with "value="
            stringIn = stringIn.replace("self.testcaseDataDict", "value=self.testcaseDataDict")
            stringIn = stringIn.replace("'), value=", "', value=")

        if "self.browserSession.findWaitNotVisible" in stringIn:
            # there's a timeout-value, that needs to be a parameter "timeout="
            # formats:
            #    <bla>,<blank><timeout>)
            #    <bla>,<timeout>)

            lFirst = stringIn.split(",")
            lEnd = lFirst[-1]
            lEnd = ", timeout = " + lEnd
            stringIn = " ".join(lFirst[0:-1]) + lEnd
            stringIn = stringIn.replace("), timeout =", ", timeout =")

        stringIn = stringIn.replace("else: if", "elif")

        # Fix space vs. tab in the file beginning
        lengthDiffSpace = len(stringIn) - len(stringIn.lstrip(' '))
        if lengthDiffSpace > 0:
            if lengthDiffSpace >= 8:
                stringIn = '\t\t' + stringIn.lstrip(' ')
            elif lengthDiffSpace >= 4:
                stringIn = '\t' + stringIn.lstrip(' ')
        return stringIn


    @staticmethod
    def _replaceGlobalVariable(stringIn):
        """
        We have any of the following occurances:
        covered: <blank>GlobalVariable.variablename<blank>
        covered: (GlobalVariable.variablename)
        covered: (GlobalVariable.variablename<blank>
        covered: <blank>GlobalVariable.variablename<lineend>
        covered: <blank>GlobalVariable.variablename)
        covered: <blank>GlobalVariable.variablename.contains('<text>'):
        <blank>GlobalVariable.variablename.toLowerCase()<blank>
        covered: <linestart>GlobalVariable.variablename<blank>

        That was the old code:
        #l_string = l_string.replace("((GlobalVariable.", "self.testcaseDataDict['")
        #l_string = l_string.replace("GlobalVariable.", "self.testcaseDataDict['")
        #l_string = l_string.replace(" ==", "'] ==")
        #l_string = l_string.replace(" !=", "'] !=")
        """
        if not "GlobalVariable" in stringIn:
            return stringIn

        while "GlobalVariable" in stringIn:
            addCRLF = False
            replaceVariable = stringIn.split("GlobalVariable")[1]
            replaceVariable = replaceVariable.split(" ")[0]
            replaceVariable = replaceVariable.split(")")[0]
            replaceVariable = replaceVariable.split(".")[1]
            if "\n" in replaceVariable:
                replaceVariable = replaceVariable.split("\n")[0]
            destination = "self.testcaseDataDict['" + replaceVariable + "']"
            replaceVariable = "GlobalVariable." + replaceVariable
            stringIn = stringIn.replace(replaceVariable, destination)
            if addCRLF:
                stringIn = stringIn + "\n"

        if ".length()" in stringIn:
            stringIn = translateGoovy._replaceLengthMethodGroovy(stringIn)
        if ".toLowerCase()" in stringIn:
            stringIn = stringIn.replace("toLowerCase()", "lower()")
        return stringIn
        pass

    @staticmethod
    def _replaceLengthMethodGroovy(stringIn):
        """bla.length() in Groovy should become len(bla)"""
        # Find the variable, that we want to know the length of:
        stringIn = stringIn.strip()
        partsBeforeAndAfter = stringIn.split(".length()")
        variableParts = partsBeforeAndAfter[0].split(" ")
        variableParts[-1] = "len(" + variableParts[-1]
        partsBeforeAndAfter[0] = " ".join(variableParts) + ")"
        return "".join(partsBeforeAndAfter)

class translateObjectDefinition(fileHandling):
    def __init__(self, fileNameAndPath):
        super().__init__(fileNameAndPath)
        self.selector = None
        self.selectorValue = None
        self.iframe = None
        self.analyze()
        self.outputAnalysis()

    def outputAnalysis(self):
        super().logFileContentsHeader()
        logger.debug(f"Selektor: {self.selector}")
        logger.debug(f"Value: {self.selectorValue}")
        logger.debug(f"Iframe: {self.iframe}")

    def analyze(self):
        nameTag = self.fileAsDOM.find("name")
        self.name = nameTag.text
        self.selector = self.fileAsDOM.find("selectormethod").get_text(strip=True) # XPATH, CSS
        selectorDrillDown = self.fileAsDOM.findAll("selectorcollection")
        for node in selectorDrillDown:
            lEntries = node.findAll("entry")
            for entry in lEntries:
                if entry.find("key").get_text(strip=True) == self.selector:
                    self.selectorValue = entry.find("value").get_text(strip=True)

        # Find potential Iframe:
        iframeDrilldown = self.fileAsDOM.find("webelementproperties")
        if iframeDrilldown:
            if iframeDrilldown.find("name").get_text(strip=True) == "ref_element":
                self.iframe = iframeDrilldown.find("value").get_text(strip=True)


def doImport(importDir):
    excludeDirs = ["Check Points", "Test Suites", ".git", "Reports", "/bin", "/settings", "/Drivers", "/Plugins",
                   "/Data Files", "/0testdateninput", "/2shared", "/1testout", "/Keywords", "/report", "/Libs",
                   "/Checkpoints", "/Include", "/3tools", "/Scripts", "/Test Listen"]
    for (root, dirs, files) in os.walk(importDir, topdown=True):
        lContinue = True
        for excludeDir in excludeDirs:
            if excludeDir in root:
                lContinue = False
                break
        if not lContinue:
            continue
        logger.info(str(root) + str(dirs) + str(files))
        for file in files:
            fileAndPath = _getFileAndPath(file=file, path=root)
            #if file[-3:] == '.rs':
            #    lDefinition = translateObjectDefinition(fileAndPath)
            #    locatorObjects.add(lDefinition, fileAndPath)
            if file[-3:] == '.tc':
                lTestCase = translateTestCase(fileAndPath)
                testCases.add(lTestCase, fileAndPath)
            #if file[-7:] == '.groovy':
            #    lGroovy = translateGoovy(fileAndPath)
            #    groovy.add(lGroovy, fileAndPath)

def _getFileAndPath(file, path):
    return path + "/" + file

def exportResults():
    lXLS = xlsxwriter.Workbook("export.xlsx")
    lWSObjects = lXLS.add_worksheet("Locators")
    locatorObjects.exportXLS(lWSObjects)
    lXLS.close()


if __name__ == '__main__':
    logger = setupLogger()
    IMPORTDIR = '/Users/bernhardbuhl/git/KatalonVIG'
    locatorObjects = LocatorObjects()
    testCases = TestCases()
    groovy = Groovy()
    doImport(IMPORTDIR)
    groovy.doReplacementOfLiterals(objects=locatorObjects)
    exportResults()
    print("Python-Files created in Katalon-Basepath/Python")
    print("Excel Export of Locators in export.xlsx")


from baangt.base.ApiHandling import ApiHandling

class SOAPHandling(ApiHandling):
    """
    This is the Class to handle SOAP-Specific functionality of baangt. It inherits all methods and parameters from
    ApiHandling
    """
    pass




import baangt

from baangt.base.BrowserHandling.BrowserHandling import BrowserDriver
from baangt.base import GlobalConstants as GC
import logging

logger = logging.getLogger("pyC")


class BrowserDriverHookImpl:
    @baangt.hook_impl
    def browserDriver_init(self, timing=None, screenshotPath=None):
        return BrowserDriver(timing, screenshotPath)

    @baangt.hook_impl
    def browserDriver_createNewBrowser(self, browserDriverObject, browserName=GC.BROWSER_FIREFOX,
                                       desiredCapabilities=None, **kwargs):
        return browserDriverObject.createNewBrowser(browserName=browserName,
                                                    desiredCapabilities=desiredCapabilities, **kwargs)

    @baangt.hook_impl
    def browserDriver_slowExecutionToggle(self, browserDriverObject, newSlowExecutionWaitTimeInSeconds=None):
        return browserDriverObject.slowExecutionToggle(newSlowExecutionWaitTimeInSeconds)

    @baangt.hook_impl
    def browserDriver_closeBrowser(self, browserDriverObject):
        return browserDriverObject.closeBrowser()

    @baangt.hook_impl
    def browserDriver_takeScreenshot(self, browserDriverObject, screenShotPath=None):
        return browserDriverObject.takeScreenshot(screenShotPath)

    @baangt.hook_impl
    def browserDriver_handleIframe(self, browserDriverObject, iframe=None):
        return browserDriverObject.handleIframe(iframe)

    @baangt.hook_impl
    def browserDriver_handleWindow(self, browserDriverObject, windowNumber=None, function=None):
        return browserDriverObject.handleWindow(windowNumber, function)

    @baangt.hook_impl
    def browserDriver_findByAndWaitForValue(self, browserDriverObject, id=None, css=None, xpath=None, class_name=None,
                                            iframe=None, timeout=20,
                                            optional=False):
        return browserDriverObject.findByAndWaitForValue(id, css, xpath, class_name,
                                            iframe, timeout,
                                            optional)

    @baangt.hook_impl
    def browserDriver_findByAndSetText(self, browserDriverObject, id=None, css=None, xpath=None, class_name=None,
                                       value=None, iframe=None,
                                       timeout=60, optional=False):
        return browserDriverObject.findByAndSetText(id, css, xpath, class_name, value,
                                                    iframe, timeout,
                                                    optional)

    @baangt.hook_impl
    def browserDriver_findByAndSetTextIf(self, browserDriverObject, id=None, css=None, xpath=None, class_name=None,
                                         value=None, iframe=None,
                                         timeout=60):
        return browserDriverObject.findByAndSetTextIf(id, css, xpath, class_name,
                                         value, iframe,
                                         timeout)

    @baangt.hook_impl
    def browserDriver_findByAndSetTextValidated(self, browserDriverObject, id=None,
                                                css=None,
                                                xpath=None,
                                                class_name=None,
                                                value=None,
                                                iframe=None,
                                                timeout=60,
                                                retries=5):
        return browserDriverObject.findByAndSetTextValidated(id,
                                                css,
                                                xpath,
                                                class_name,
                                                value,
                                                iframe,
                                                timeout,
                                                retries)

    @baangt.hook_impl
    def browserDriver_submit(self, browserDriverObject):
        return browserDriverObject.submit()

    @baangt.hook_impl
    def browserDriver_findByAndClick(self, browserDriverObject, id=None, css=None, xpath=None, class_name=None,
                                     iframe=None, timeout=20, optional=False):
        return browserDriverObject.findByAndClick(id, css, xpath, class_name,
                                     iframe, timeout, optional)

    @baangt.hook_impl
    def browserDriver_findByAndClickIf(self, browserDriverObject, id=None, css=None, xpath=None, class_name=None,
                                       iframe=None, timeout=60,
                                       value=None, optional=False):
        return browserDriverObject.findByAndClickIf(id, css, xpath, class_name,
                                       iframe, timeout,
                                       value, optional)

    @baangt.hook_impl
    def browserDriver_findByAndForceText(self, browserDriverObject, id=None, css=None, xpath=None, class_name=None,
                                         value=None,
                                         iframe=None, timeout=60, optional=False):
        return browserDriverObject.findByAndForceText(id, css, xpath, class_name,
                                         value,
                                         iframe, timeout, optional)

    @baangt.hook_impl
    def browserDriver_findBy(self, browserDriverObject, id=None, css=None, xpath=None, class_name=None, iframe=None,
                             timeout=60, loggingOn=True,
                             optional=False):
        return browserDriverObject.findBy(id, css, xpath, class_name, iframe,
                             timeout, loggingOn,
                             optional)

    @baangt.hook_impl
    def browserDriver_getURL(self, browserDriverObject):
        return browserDriverObject.getURL()

    @baangt.hook_impl
    def browserDriver_findWaitNotVisible(self, browserDriverObject, xpath=None, id=None, timeout=90):
        return browserDriverObject.findWaitNotVisible(xpath, id, timeout)

    @baangt.hook_impl
    def browserDriver_sleep(self, browserDriverObject, sleepTimeinSeconds):
        return browserDriverObject.sleep(sleepTimeinSeconds)

    @baangt.hook_impl
    def browserDriver_goToUrl(self, browserDriverObject, url):
        return browserDriverObject.goToUrl(url)

    @baangt.hook_impl
    def browserDriver_goBack(self, browserDriverObject):
        return browserDriverObject.goBack()

    @baangt.hook_impl
    def browserDriver_javaScript(self, browserDriverObject, jsText):
        return browserDriverObject.javaScript(jsText)


